"""
Neuraxle's Hyperparameter Repository Base Classes
====================================================

Not all repositories can be found here, but the base can be found here.


.. seealso::
    :class:`~neuraxle.metaopt.hyperopt.repositories.json.HyperparamsOnDiskRepository`,
    :class:`~neuraxle.metaopt.hyperopt.repositories.db.DatabaseHyperparamRepository`,
    :class:`~neuraxle.metaopt.hyperopt.repositories.db.SQLLiteHyperparamsRepository`,


..
    Copyright 2022, Neuraxio Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.


"""

import copy
import multiprocessing
import os
from abc import abstractmethod

from neuraxle.base import BaseService, MetaServiceMixin
from neuraxle.logging.logging import NeuraxleLogger
from neuraxle.logging.warnings import RaiseDeprecatedClass
from neuraxle.metaopt.data.vanilla import (BaseDataclass, RootDataclass,
                                           ScopedLocation, ScopedLocationAttr,
                                           SubDataclassT,
                                           dataclass_2_subloc_attr)


class HyperparamsRepository(BaseService):
    """
    Hyperparams repository that saves hyperparams, and scores for every AutoML trial.
    Cache folder can be changed to do different round numbers.

    For more information, read this `article by Martin Fowler on DDD Aggregates <https://martinfowler.com/bliki/DDD_Aggregate.html>`_.

    .. seealso::
        :class:`AutoML`,
        :class:`Trainer`,
    """

    def __init__(self):
        BaseService.__init__(self)

    @abstractmethod
    def load(self, scope: ScopedLocation, deep=False) -> SubDataclassT:
        """
        Get metadata from scope.

        The fetched metadata will be the one that is the last item
        that is not a None in the provided scope.

        :param scope: scope to get metadata from.
        :return: metadata from scope.
        """
        raise NotImplementedError("Use a concrete class. This is an abstract class.")

    @abstractmethod
    def save(self, _dataclass: SubDataclassT, scope: ScopedLocation, deep=False) -> 'HyperparamsRepository':
        """
        Save metadata to scope.

        :param metadata: metadata to save.
        :param scope: scope to save metadata to.
        :param deep: if True, save metadata's sublocations recursively.
        """
        raise NotImplementedError("Use a concrete class. This is an abstract class.")

    @abstractmethod
    def add_logging_handler(self, logger: NeuraxleLogger, scope: ScopedLocation) -> 'HyperparamsRepository':
        """
        Add logging handler to repository's provided scope.
        Handler must be set manually for each scope below this scope.

        :param logger: logger to add handler to.
        :param scope: scope to add handler to.
        :return: self.
        """
        raise NotImplementedError("Use a concrete class. This is an abstract class.")

    @abstractmethod
    def get_log_from_logging_handler(self, logger: NeuraxleLogger, scope: ScopedLocation) -> str:
        """
        Get log from repository's provided scope and handler that was set with :func:`add_logging_handler`.

        :param scope: scope to get log from.
        :return: log from scope.
        """
        raise NotImplementedError("Use a concrete class. This is an abstract class.")

    def with_lock(self) -> 'SynchronizedHyperparamsRepositoryWrapper':
        return SynchronizedHyperparamsRepositoryWrapper(self)


def func_with_rlock():
    def decorator(func):
        def _LOCKED_REPO(self, *args, **kwargs):
            with self.lock:
                return func(self, *args, **kwargs)
        return _LOCKED_REPO
    return decorator


class SynchronizedHyperparamsRepositoryWrapper(MetaServiceMixin, HyperparamsRepository):
    """
    A wrapper that makes any HyperparamsRepository thread-safe using locking.
    """

    def __init__(self, wrapped: HyperparamsRepository):
        MetaServiceMixin.__init__(self, wrapped)
        HyperparamsRepository.__init__(self)
        self.lock = multiprocessing.RLock()

    @func_with_rlock()
    def load(self, scope: ScopedLocation, deep=False) -> SubDataclassT:
        repo: HyperparamsRepository = self.get_step()
        return repo.load(scope, deep)

    @func_with_rlock()
    def save(self, _dataclass: SubDataclassT, scope: ScopedLocation, deep=False) -> 'HyperparamsRepository':
        repo: HyperparamsRepository = self.get_step()
        return repo.save(_dataclass, scope, deep)

    @func_with_rlock()
    def add_logging_handler(self, logger: NeuraxleLogger, scope: ScopedLocation) -> 'HyperparamsRepository':
        repo: HyperparamsRepository = self.get_step()
        return repo.add_logging_handler(logger, scope)

    @func_with_rlock()
    def get_log_from_logging_handler(self, logger: NeuraxleLogger, scope: ScopedLocation) -> str:
        repo: HyperparamsRepository = self.get_step()
        return repo.get_log_from_logging_handler(logger, scope)

#     def _get_thread_safe(self):
#         pass
#
#     def _get_pickle_safe(self):
#         pass
#
#     def set_thread_safe():
#         pass

    def with_lock(self) -> 'SynchronizedHyperparamsRepositoryWrapper':
        return self


class _InMemoryRepositoryLoggerHandlerMixin:
    """
    Mixin to add a in-memory logging handler to a repository.
    """

    def add_logging_handler(self, logger: NeuraxleLogger, scope: ScopedLocation) -> 'HyperparamsRepository':
        return self

    def get_log_from_logging_handler(self, logger: NeuraxleLogger, scope: ScopedLocation) -> str:
        return logger.get_scoped_string_history()


class VanillaHyperparamsRepository(_InMemoryRepositoryLoggerHandlerMixin, HyperparamsRepository):
    """
    Hyperparams repository that saves data AutoML-related info.
    """

    def __init__(
        self,
        cache_folder: str
    ):
        """
        :param cache_folder: folder to store trials.
        :param hyperparams_repo_class: class to use to save hyperparams.
        :param hyperparams_repo_kwargs: kwargs to pass to hyperparams_repo_class.
        """
        HyperparamsRepository.__init__(self)
        _InMemoryRepositoryLoggerHandlerMixin.__init__(self)
        self.cache_folder = os.path.join(cache_folder, self.__class__.__name__)
        self.root: RootDataclass = RootDataclass()

    @staticmethod
    def from_root(root: RootDataclass, cache_folder: str) -> 'VanillaHyperparamsRepository':

        return VanillaHyperparamsRepository(
            cache_folder=cache_folder,
        ).save(root, ScopedLocation(), deep=True)

    def load(self, scope: ScopedLocation, deep=False) -> SubDataclassT:
        try:
            ret: BaseDataclass = self.root[scope]
            if not deep:
                ret = ret.shallow()
        except KeyError:
            ret: BaseDataclass = scope.new_dataclass_from_id()

        return copy.deepcopy(ret)

    def save(self, _dataclass: SubDataclassT, scope: ScopedLocation, deep=False) -> 'VanillaHyperparamsRepository':
        # Sanitizing
        _dataclass: SubDataclassT = copy.deepcopy(_dataclass)
        scope = scope.at_dc(_dataclass)

        # Sanity checks: good type
        if not isinstance(_dataclass, BaseDataclass):
            raise ValueError(f"Was expecting a dataclass. Got `{_dataclass.__class__.__name__}`.")
        # Sanity checks: sufficient scope depth
        scope = scope[:_dataclass.__class__]  # Sanitizing scope to dtype loc.
        _id_from_scope: ScopedLocationAttr = scope[_dataclass.__class__]
        if _id_from_scope is not None:
            if _id_from_scope != _dataclass.get_id():
                raise ValueError(
                    f"The scope `{scope}` with {_dataclass.__class__.__name__} id `{_id_from_scope}` does not match the provided dataclass id `{_dataclass.get_id()}` for `{_dataclass}`."
                )
            scope.pop()
            # else check if the scope is at least of the good class length:
        elif len(scope) != list(dataclass_2_subloc_attr.keys()).index(_dataclass.__class__):
            raise ValueError(
                f"The scope `{scope}` is not of the good length for dataclass type `{_dataclass.__class__.__name__}`."
            )

        # Passthrough dc's sublocation when saving shallow:
        if not deep:
            if isinstance(_dataclass, RootDataclass):
                _dataclass.set_sublocation(self.root.get_sublocation())
            elif scope.with_dc(_dataclass) in self.root:
                prev_dc: SubDataclassT = self.root[scope.with_dc(_dataclass)]
                _dataclass.set_sublocation(prev_dc.get_sublocation())
            else:
                _dataclass = _dataclass.empty()

        # Finally storing.
        if isinstance(_dataclass, RootDataclass):
            self.root = _dataclass
        else:
            self.root[scope].store(_dataclass)
        return self


class InMemoryHyperparamsRepository(RaiseDeprecatedClass, VanillaHyperparamsRepository):
    """
    In memory hyperparams repository that can print information about trials.
    Useful for debugging.
    """

    def __init__(self, *kargs, **kwargs):
        RaiseDeprecatedClass.__init__(
            self,
            replacement_class=VanillaHyperparamsRepository,
            since_version="0.7.0",
        )
        VanillaHyperparamsRepository.__init__(self, *kargs, **kwargs)



"""
Neuraxle's DataContainer classes
====================================
Classes for containing the data that flows throught the pipeline steps.

..
    Copyright 2019, Neuraxio Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

..
    Thanks to Umaneo Technologies Inc. for their contributions to this Machine Learning
    project, visit https://www.umaneo.com/ for more information on Umaneo Technologies Inc.

"""
import copy
import math
from operator import attrgetter
from typing import Any, Callable, Generic, Iterable, Iterator, List, Optional, Tuple, TypeVar, Union

import numpy as np
import pandas as pd

NamedDACTTuple = Tuple[str, 'DataContainer']
IDT = TypeVar('IDT', bound=Iterable)  # Ids Type that is often a list of things
DIT = TypeVar('DIT', bound=Iterable)  # Data Inputs Type that is often a list of things
EOT = TypeVar('EOT', bound=Iterable)  # Expected Outputs Type that is often a list of things
DACTData = Union[IDT, DIT, EOT]  # Any of the 3 types

ARG_X_INPUTTED = DIT
ARG_Y_EXPECTED = EOT
ARG_Y_PREDICTD = DIT


class StripAbsentValues:
    """
    This object, when passed to the default_value_data_inputs argument of the DataContainer.batch method,
    will return the minibatched data containers such that the last batch won't have the full batch_size
    if it was incomplete with trailing None values at the end.
    """
    pass


class DataContainer(Generic[IDT, DIT, EOT]):
    """
    DataContainer (dact) class to store IDs (ids), data inputs (di), and expected outputs (eo) together.
    In some dacts, you could have only ids and data inputs, and in other dacts you could have only expected outputs,
    or you could have all if you want, such as when your :class:`~neuraxle.pipeline.Pipeline` is used to train a model
    in a certain :class:`~neuraxle.base.ExecutionMode` within a certain :class:`~neuraxle.base.ExecutionContext`.

    You can use typing for your dact, and create a dact, such as:

    .. code-block:: python

        from typing import List
        from neuraxle.data_container import DataContainer as DACT

        dact: DACT[List[str], List[int], List[float]] = DACT(
            ids=['a', 'b', 'c'],
            data_inputs=[1, 2, 3],
            expected_outputs=[1.0, 2.0, 3.0]
        )


    This is because the DataContainer inherits from the :class:`~typing.Generic` type
    as ``class DataContainer(Generic[IDT, DIT, EOT]): ...``.

    The DataContainer object is passed to all of the :class:`~neuraxle.base.BaseStep` 's handle methods :
        * :func:`~neuraxle.base.BaseStep.handle_transform`
        * :func:`~neuraxle.base.BaseStep.handle_fit_transform`
        * :func:`~neuraxle.base.BaseStep.handle_fit`

    Most of the time, the steps will manage it in the handler methods.

    .. seealso::
        :class:`~neuraxle.base.BaseStep`,
        :class:`~neuraxle.data_container.StripAbsentValues`
    """

    def __init__(
            self,
            data_inputs: Optional[DIT] = None,
            ids: Optional[IDT] = None,
            expected_outputs: Optional[EOT] = None,
            sub_data_containers: List['NamedDACTTuple'] = None,
            *,
            di: Optional[DIT] = None,
            eo: Optional[EOT] = None,

    ):
        """
        Create a DataContainer[IDT, DIT, EOT] object from specified ids, di, and eo.

        :param ids: ids that are iterable. If None, will put a range of integers of data_inputs length. Often a list of integers.
        :param di: same as ``data_inputs``, but shorter.
        :param eo: same as ``expected_outputs``, but shorter.
        :param data_inputs: data inputs that are iterable. Can use di instead.
        :param expected_outputs: expected outputs that are iterable. If None, will put a list of None of data_inputs length.
        :param sub_data_containers: sub data containers.
        """
        self._ids: IDT = ids
        self.data_inputs: DIT = data_inputs if di is None else di
        self.expected_outputs: EOT = expected_outputs if eo is None else eo
        self.sub_data_containers: List[NamedDACTTuple] = sub_data_containers or []

    @property
    def ids(self) -> IDT:
        """
        Get ids.

        If the ids are None, the following IDs will be returned:

        - If the data_inputs is a :class:`~pandas.DataFrame`, will return the index of the DF.
        - Else if the ids are None, will return a range of integers of data_inputs length.
        - Else if the data_inputs aren't iterable, will return a range of integers of expected_outputs length.

        :return: ids
        :rtype: Iterable
        """
        if self._ids is None:
            if hasattr(self.di, 'index') and hasattr(self.di.index, 'values'):
                # This is a pd.DataFrame:
                return self.di.index.values
            if hasattr(self.di, '__len__') or hasattr(self.di, '__iter__'):
                return [i for i in range(len(self.di))]
            if hasattr(self.eo, '__len__') or hasattr(self.eo, '__iter__'):
                return [i for i in range(len(self.eo))]
        return self._ids

    @ids.setter
    def ids(self, ids: IDT):
        """
        Set ids.

        :param ids: ids, often a list of integers.
        :type ids: Iterable
        """
        self._ids = ids

    @property
    def di(self) -> DIT:
        """
        Get data inputs.

        :return: data inputs
        :rtype: Iterable
        """
        return self.data_inputs

    @di.setter
    def di(self, di: DIT):
        """
        Set data inputs.

        :param di: data inputs
        :type di: Iterable
        """
        self.data_inputs = di

    @property
    def eo(self) -> EOT:
        """
        Get expected outputs.

        If the expected outputs are None, will return a list of None of data_inputs length.

        :return: expected outputs
        :rtype: Iterable
        """
        if self.expected_outputs is None and (
            hasattr(self.data_inputs, '__len__') or hasattr(self.data_inputs, '__iter__')
        ):
            return [None] * len(self.data_inputs)
        return self.expected_outputs

    @eo.setter
    def eo(self, eo: EOT):
        """
        Set expected outputs.

        :param eo: expected outputs
        :type eo: Iterable
        """
        self.expected_outputs = eo

    @property
    def sdact(self) -> List['NamedDACTTuple']:
        """
        Get sub data containers.

        :return: sub data containers
        """
        return self.sub_data_containers

    @staticmethod
    def from_di(data_inputs: DIT) -> 'DACT[IDT, DIT, List[None]]':
        """
        Create a DataContainer (dact) from data inputs (di).
        """
        return DACT(
            ids=None,
            di=data_inputs,
            eo=None
        )

    @staticmethod
    def from_eo(expected_outputs: EOT) -> 'DACT[List[None], List[None], EOT]':
        """
        Create a DataContainer (dact) from expected outputs (eo).
        """
        return DACT(
            ids=None,
            di=None,
            eo=expected_outputs
        )

    def without_di(self) -> 'DACT[IDT, List[None], EOT]':
        return self.copy().set_data_inputs(None)

    def without_eo(self) -> 'DACT[IDT, DIT, List[None]]':
        return self.copy().set_expected_outputs(None)

    def with_di(self, di: DIT) -> 'DACT[IDT, DIT, EOT]':
        return self.copy().set_data_inputs(di)

    def with_eo(self, eo: EOT) -> 'DACT[IDT, DIT, EOT]':
        return self.copy().set_expected_outputs(eo)

    def with_ids(self, ids: DIT) -> 'DACT[IDT, DIT, EOT]':
        return self.copy().set_ids(ids)

    def set_ids(self, ids: IDT) -> 'DACT':
        """
        Set ids.

        :param ids: data inputs' ids. Often a range of integers.
        :return: self
        """
        self.ids = ids
        return self

    def set_data_inputs(self, data_inputs: DIT) -> 'DACT':
        """
        Set data inputs.

        :param data_inputs: data inputs
        :type data_inputs: Iterable
        :return: self
        """
        self.data_inputs = data_inputs
        return self

    def set_expected_outputs(self, expected_outputs: EOT) -> 'DACT':
        """
        Set expected outputs.

        :param expected_outputs: expected outputs
        :type expected_outputs: Iterable
        :return: self
        """
        self.expected_outputs: EOT = expected_outputs
        return self

    def get_ids_summary(self) -> Optional[str]:
        if self._ids is None:
            return None
        return ','.join([str(i) for i in self.ids if i is not None])

    def add_sub_data_container(self, name: str, data_container: 'DACT') -> 'DACT':
        """
        Get sub data container if item is str, otherwise get a zip of ids, data inputs, and expected outputs.

        :type name: sub data container name
        :type data_container: sub data container
        :return: self
        """
        self.sub_data_containers.append((name, data_container))
        return self

    def get_sub_data_container_names(self) -> List[str]:
        """
        Get sub data container names.

        :return: list of names
        """
        return [name for name, _ in self.sub_data_containers]

    def set_sub_data_containers(self, sub_data_containers: List['DACT']) -> 'DACT':
        """
        Set sub data containers
        :return: self
        """
        self.sub_data_containers = sub_data_containers
        return self

    def minibatches(
            self,
            batch_size: int,
            keep_incomplete_batch: bool = True,
            default_value_data_inputs=None,
            default_value_expected_outputs=None
    ) -> Iterable['DACT[IDT, DIT, EOT]']:
        """
        Yields minibatches extracted from looping on the DataContainer's content with a batch_size and a certain behavior for the last batch when the batch_size is uneven with the total size.

        Note that the default value for IDs is None.

        .. code-block:: python

            data_container = DataContainer(data_inputs=np.array(list(range(10)))
            for data_container_batch in data_container.minibatches(batch_size=2):
                print(data_container_batch.data_inputs)
                print(data_container_batch.expected_outputs)
            # [array([0, 1]), array([2, 3]), ..., array([8, 9])]

            data_container = DataContainer(data_inputs=np.array(list(range(10)))
            for data_container_batch in data_container.minibatches(batch_size=3, keep_incomplete_batch=False):
                print(data_container_batch.data_inputs)
            # [array([0, 1, 2]), array([3, 4, 5]), array([6, 7, 8])]

            data_container = DataContainer(data_inputs=np.array(list(range(10)))
            for data_container_batch in data_container.minibatches(
                batch_size=3,
                keep_incomplete_batch=True,
                default_value_data_inputs=None,
                default_value_expected_outputs=None
            ):
                print(data_container_batch.data_inputs)
            # [array([0, 1, 2]), array([3, 4, 5]), array([6, 7, 8]), array([9, None, None])]

            data_container = DataContainer(data_inputs=np.array(list(range(10)))
            for data_container_batch in data_container.minibatches(
                batch_size=3,
                keep_incomplete_batch=True,
                default_value_data_inputs=StripAbsentValues()
            ):
                print(data_container_batch.data_inputs)
            # [array([0, 1, 2]), array([3, 4, 5]), array([6, 7, 8]), array([9])]


        :param batch_size: number of elements to combine into a single batch
        :param keep_incomplete_batch: (Optional.) A bool representing
        whether the last batch should be dropped in the case it has fewer than
        `batch_size` elements; the default behavior is to keep the smaller
        batch.
        :param default_value_data_inputs: expected_outputs default fill value
        for padding and values outside iteration range, or :class:`~neuraxle.data_container.StripAbsentValues`
        to trim absent values from the batch
        :param default_value_expected_outputs: expected_outputs default fill value
        for padding and values outside iteration range, or :class:`~neuraxle.data_container.StripAbsentValues`
        to trim absent values from the batch
        :return: an iterator of DataContainer

        .. seealso::
            :class:`~neuraxle.data_container.StripAbsentValues`
        """
        for i in range(0, len(self.data_inputs), batch_size):
            data_container: DACT[IDT, DIT, EOT] = DACT(
                ids=self.ids[i:i + batch_size] if self._ids is not None else None,
                data_inputs=self.di[i:i + batch_size] if self.data_inputs is not None else None,
                expected_outputs=self.eo[i:i + batch_size] if self.expected_outputs is not None else None
            )

            incomplete_batch = len(data_container.data_inputs) < batch_size
            if incomplete_batch:
                if not keep_incomplete_batch:
                    break

                data_container = _pad_or_keep_incomplete_batch(
                    data_container,
                    batch_size,
                    default_value_data_inputs,
                    default_value_expected_outputs
                )

            yield data_container

    def get_n_batches(self, batch_size: int, keep_incomplete_batch: bool = True) -> int:
        if keep_incomplete_batch:
            return math.ceil(len(self.data_inputs) / batch_size)
        else:
            return math.floor(len(self.data_inputs) / batch_size)

    def copy(self) -> 'DACT[IDT, DIT, EOT]':
        return DACT(
            ids=self._ids,
            di=self.data_inputs,
            eo=self.expected_outputs,
            sub_data_containers=[(name, data_container.copy()) for name, data_container in self.sub_data_containers]
        )

    def __contains__(self, item: Union[str, int]) -> bool:
        """
        return true if sub container name is in the sub data containers, or if has ids of the good number.

        :param item: sub container name if string, id if int.
        :return: contains
        """
        if isinstance(item, str):
            contains = False
            for name, data_container in self.sub_data_containers:
                if name == item:
                    contains = True
            return contains
        else:
            if item in self.ids:
                return True
            else:
                return False
        raise NotImplementedError('DataContainer.__contains__ not implemented for this type: {}'.format(type(item)))

    def __getitem__(self, item: Union[str, int]) -> Union['DACT', Tuple[IDT, DIT, EOT]]:
        """
        If item is a string, then get then sub container with the same name as item in the list of sub data containers.
        If item is an int, then return a tuple of (id, data input, expected output) for the given item index.

        :param item: sub data container str, or item index as int
        :type item: Union[str, int]
        :return: data container, or tuple of ids, data inputs, expected outputs.
        :rtype: Union[DataContainer, Tuple]
        """
        if isinstance(item, str):
            for name, data_container in self.sub_data_containers:
                if name == item:
                    return data_container
            raise KeyError("sub_data_container {} not found in data container".format(item))
        else:
            if self.ids is None:
                ids = [None] * len(self)
            else:
                ids = self.ids[item]

            return ids, self.data_inputs[item], self.expected_outputs[item]

    def tolist(self) -> 'DACT[List, List, List]':
        def pandas_tonp(df):
            if 'DataFrame' in str(type(df)) or 'Series' in str(type(df)):
                return df.values.tolist()
            else:
                return df

        def ndarr_tolist(ndarr):
            if isinstance(ndarr, np.ndarray):
                return ndarr.tolist()
            else:
                return ndarr

        self.apply_conversion_func(pandas_tonp)
        self.apply_conversion_func(ndarr_tolist)
        return self

    def tolistshallow(self) -> 'DACT[List, List, List]':
        return self.apply_conversion_func(list)

    def to_numpy(self) -> 'DACT[np.ndarray, np.ndarray, np.ndarray]':
        return self.apply_conversion_func(np.array)

    def apply_conversion_func(self, conversion_function: Callable[[Any], Any]) -> 'DACT':
        """
        Apply conversion function to data inputs, expected outputs, and ids,
        and set the new values in self. Returns self.
        Conversion function must be able to handle None values.
        """
        self.set_ids(
            conversion_function(self.ids)
            if self.ids is not None
            else None
        )
        self.set_data_inputs(
            conversion_function(self.data_inputs)
            if self.data_inputs is not None
            else None
        )
        self.set_expected_outputs(
            conversion_function(self.expected_outputs)
            if self.expected_outputs is not None
            else None
        )
        return self

    def unpack(self) -> Tuple[IDT, DIT, EOT]:
        """
        Unpack to a tuples of (ids, data input, expected output).

        :return: tuple of ids, data inputs, expected outputs
        """
        return self.ids, self.di, self.eo

    def __iter__(self) -> Iterator[Tuple[IDT, DIT, EOT]]:
        """
        Iter method returns a zip of all of the ids, data_inputs, and expected_outputs in the data container.

        :return: iterator of tuples containing ids, data_inputs, and expected outputs
        :rtype: Iterator[Tuple]
        """
        if self.data_inputs is None:
            return iter(())

        _ids: Optional[List[DACTData]] = self.ids
        _di: Optional[List[DACTData]] = self.di
        _eo: Optional[List[DACTData]] = self.eo
        if _ids is None or _di is None or _eo is None:
            return iter(())

        return zip(_ids, _di, _eo)

    def __repr__(self):
        return str(self)

    def __str__(self):
        ids = self._ids
        di = self.data_inputs
        eo = self.expected_outputs
        ids_rep = self._str_data(ids)
        di_rep = self._str_data(di)
        eo_rep = self._str_data(eo)
        return (
            f"{self.__class__.__name__}[{type(ids).__name__}, {type(di).__name__}, {type(eo).__name__}](\n"
            f"\tids={ids_rep},\n"
            f"\tdi={di_rep},\n"
            f"\teo={eo_rep}\n)"
        )

    def _str_data(self, _idata: DACTData) -> str:
        if _idata is None:
            return str(None)

        if len(_idata) > 10 and hasattr(_idata, '__getitem__'):
            _shortrepr = repr(_idata[:15])
        else:
            _shortrepr = repr(_idata)
        _shortrepr = _shortrepr[:70] + ("" if len(_shortrepr) < 70 else "...")

        _len = "len=?"
        if isinstance(_idata, pd.DataFrame):
            _len = f"shape={_idata.values.shape}"
        elif isinstance(_idata, np.ndarray):
            _len = f"shape={_idata.shape}"
        elif hasattr(_idata, "__len__"):
            _len = f"len={len(_idata)}"

        _len_rep = f"<`{_shortrepr}` of {_len}>"
        return _len_rep.replace("\n", " ").replace("\t", " ").replace("    ", " ")

    def __len__(self):
        return len(self.data_inputs)


DACT = DataContainer
TrainDACT = DACT[IDT, ARG_X_INPUTTED, ARG_Y_EXPECTED]  # training set input
ValidDACT = DACT[IDT, ARG_X_INPUTTED, ARG_Y_EXPECTED]  # validation set input
PredsDACT = DACT[IDT, ARG_Y_PREDICTD, Optional[EOT]]  # output after prediction

EvalEOTDACT = DACT[IDT, ARG_Y_PREDICTD, ARG_Y_EXPECTED]  # a merge of ValidDACT and PredsDACT: PRED Y and EXPECTED Y


class ExpandedDataContainer(DACT):
    """
    Sub class of DataContainer to expand data container dimension.
    This is akin from passing from `shape` to `[1, *shape]`
    when using :func:`ExpandedDataContainer.create_from`.

    .. seealso::
        :class:`DataContainer`
    """

    def __init__(self, data_inputs, ids, expected_outputs, _old_ids):
        DACT.__init__(
            self,
            ids=ids,
            di=data_inputs,
            eo=expected_outputs,
        )

        self._old_ids = _old_ids

    @staticmethod
    def create_from(data_container: DACT) -> 'ExpandedDataContainer':
        """
        Create ExpandedDataContainer with a summary id for the new id.
        This is akin from passing from `shape` to `[1, *shape]`.

        :param data_container: data container to transform
        :type data_container: DataContainer
        :return: expanded data container
        """
        return ExpandedDataContainer(
            ids=[data_container.get_ids_summary()],
            data_inputs=[data_container.data_inputs] if data_container.data_inputs is not None else None,
            expected_outputs=[data_container.expected_outputs] if data_container.expected_outputs is not None else None,
            _old_ids=data_container._ids
        )

    def reduce_dim(self) -> 'DACT':
        """
        Reduce DataContainer to its original shape with a list of multiple ids, data_inputs, and expected outputs.

        :return: reduced data container
        :rtype: DataContainer
        """
        if len(self.data_inputs) != 1:
            raise ValueError(
                'Invalid Expanded Data Container. Please create ExpandedDataContainer with ExpandedDataContainer.create_from(data_container) method.')

        return DACT(
            ids=self._old_ids,
            data_inputs=self.data_inputs[0] if self.data_inputs is not None else None,
            expected_outputs=self.expected_outputs[0] if self.expected_outputs is not None else None,
            sub_data_containers=self.sub_data_containers
        )


class ZipDataContainer(DACT):
    """
    Sub class of DataContainer to zip two data sources together.

    .. seealso::
        :class: `DataContainer`
    """

    @staticmethod
    def create_from(data_container: DACT, *other_data_containers: List[DACT], zip_expected_outputs: bool = False) -> 'ZipDataContainer':
        """
        Merges two data sources together. Zips only the data input part and keeps the expected output of the first DataContainer as is.
        NOTE: Expects that all DataContainer are at least as long as data_container.

        :param data_container: the main data container, the attribute of this data container will be kept by the returned ZipDataContainer.
        :type data_container: DataContainer
        :param other_data_containers: other data containers to zip with data container
        :type other_data_containers: List[DataContainer]
        :param zip_expected_outputs: Determines wether we kept the expected_output of data_container or we zip the expected_outputs of all DataContainer provided
        :return: expanded data container
        :rtype: ExpandedDataContainer
        """

        new_data_inputs = tuple(zip(*map(attrgetter("di"), [data_container] + list(other_data_containers))))
        if zip_expected_outputs:
            expected_outputs = tuple(
                zip(*map(attrgetter("eo"), [data_container] + list(other_data_containers))))
        else:
            expected_outputs = data_container.expected_outputs

        return ZipDataContainer(
            data_inputs=new_data_inputs,
            expected_outputs=expected_outputs,
            ids=data_container._ids,
            sub_data_containers=data_container.sub_data_containers
        )

    def concatenate_inner_features(self):
        """
        Concatenate inner features from zipped data inputs.
        Assumes each data_input entry is an iterable of numpy arrays.
        """
        new_data_inputs = [di[0] for di in self.di]

        for i, data_input in enumerate(self.di):
            new_data_inputs[i] = _inner_concatenate_np_array(list(data_input))

        self.set_data_inputs(new_data_inputs)


class ListDataContainer(DataContainer, Generic[IDT, DIT, EOT]):
    """
    Sub class of DataContainer to perform list operations.
    It allows to perform append, and concat operations on a DataContainer.

    .. seealso::
        :class:`DataContainer`
    """

    def __init__(self, data_inputs: Any, ids=None, expected_outputs: Any = None,
                 sub_data_containers=None):
        DACT.__init__(
            self,
            ids=ids,
            di=data_inputs,
            eo=expected_outputs,
            sub_data_containers=sub_data_containers
        )
        self.tolistshallow()

    @staticmethod
    def empty(original_data_container: DACT = None) -> 'ListDataContainer':
        if original_data_container is None:
            sub_data_containers = []
        else:
            sub_data_containers = original_data_container.sub_data_containers

        return ListDataContainer([], [], [], sub_data_containers=sub_data_containers)

    def append(self, _id: str, data_input: Any, expected_output: Any):
        """
        Append a new data input to the DataContainer.

        :param _id: id for the data input
        :type _id: str
        :param data_input: data input
        :param expected_output: expected output
        :return:
        """
        self._ids.append(_id)
        self.data_inputs.append(data_input)
        self.expected_outputs.append(expected_output)

    def append_data_container_in_data_inputs(self, other: DACT) -> 'ListDataContainer':
        """
        Append a data container to the data inputs of this data container.

        :param other: data container
        :type other: DataContainer
        :return:
        """
        if not isinstance(other, DACT):
            raise ValueError(
                f"Expected data container of type {DACT.__name__}, "
                f"got {other.__class__.__name__}: {other}")
        self.data_inputs.append(other)
        self._ids.append(other.get_ids_summary())
        return self

    def append_data_container(self, other: DACT) -> 'ListDataContainer':
        """
        Append a data container to the DataContainer.

        :param other: data container
        :type other: DataContainer
        :return:
        """
        self._ids.append(other._ids)
        self.data_inputs.append(other.data_inputs)
        self.expected_outputs.append(other.expected_outputs)

        return self

    def extend(self, other: DACT):
        """
        Concat the given data container at the end of self so as to extend each IDs, DIs, and EOs.

        :param data_container: data container
        :type data_container: DataContainer
        :return:
        """
        other.tolistshallow()

        if self._ids is not None and other._ids is not None:
            self._ids.extend(other._ids)
        if self.data_inputs is not None and other.data_inputs is not None:
            self.data_inputs.extend(other.data_inputs)
        if self.expected_outputs is not None and other.expected_outputs is not None:
            self.expected_outputs.extend(other.expected_outputs)

        return self


def _pad_or_keep_incomplete_batch(
        data_container,
        batch_size,
        default_value_data_inputs,
        default_value_expected_outputs
) -> DACT:
    should_pad_right = not isinstance(
        default_value_data_inputs, StripAbsentValues
    ) or not isinstance(
        default_value_expected_outputs, StripAbsentValues)

    if should_pad_right:
        data_container = _pad_incomplete_batch(
            data_container,
            batch_size,
            default_value_data_inputs,
            default_value_expected_outputs
        )

    return data_container


def _pad_incomplete_batch(
        data_container: 'DACT',
        batch_size: int,
        default_value_data_inputs: Any,
        default_value_expected_outputs: Any
) -> DACT:
    pad_di = not isinstance(default_value_data_inputs, StripAbsentValues)
    pad_eo = not isinstance(default_value_expected_outputs, StripAbsentValues)
    pad_ids = pad_di and pad_eo

    data_container = DACT(
        ids=_pad_data(
            data_container.ids,
            default_value=None,
            batch_size=batch_size
        ) if pad_ids else data_container.ids,
        data_inputs=_pad_data(
            data_container.data_inputs,
            default_value=default_value_data_inputs,
            batch_size=batch_size
        ) if pad_di else data_container.data_inputs,
        expected_outputs=_pad_data(
            data_container.expected_outputs,
            default_value=default_value_expected_outputs,
            batch_size=batch_size
        ) if pad_eo else data_container.expected_outputs,
    )

    return data_container


def _pad_data(data: DACTData, default_value: Any, batch_size: int):
    if data is None:
        return None
    data_ = []
    data_.extend(data)
    padding = copy.copy([default_value] * (batch_size - len(data)))
    data_.extend(padding)
    return data_


def _inner_concatenate_np_array(np_arrays_to_concatenate: List[np.ndarray]):
    """
    Concatenate numpy arrays on the last axis, expanding and broadcasting if necessary.

    :param np_arrays_to_concatenate: numpy arrays to zip with the other
    :type np_arrays_to_concatenate: Iterable[np.ndarray]
    :return: concatenated np array
    :rtype: np.ndarray
    """
    n_arrays = len(np_arrays_to_concatenate)
    target_n_dims = max(map(lambda x: len(x.shape), np_arrays_to_concatenate))
    # These three next lines may be a bit hard to follow;
    # they compute the max for every dimensions of the np arrays except the last.
    # Previous code just assumed np_arrays_to_concatenate[0].shape[:-1]
    _temp = map(attrgetter('shape'), np_arrays_to_concatenate)
    _temp = map(lambda x: list(x) + [0 for _ in range(target_n_dims - len(x))], _temp)
    target_dims_m1 = list(map(max, *list(_temp)))[:-1]

    for i in range(n_arrays):
        while len(np_arrays_to_concatenate[i].shape) < target_n_dims:
            np_arrays_to_concatenate[i] = np.expand_dims(np_arrays_to_concatenate[i], axis=-1)

        target_shape = tuple(target_dims_m1 + [np_arrays_to_concatenate[i].shape[-1]])
        np_arrays_to_concatenate[i] = np.broadcast_to(np_arrays_to_concatenate[i], target_shape)

    return np.concatenate(np_arrays_to_concatenate, axis=-1)



"""
Featurization Steps
==========================================================
You can find here steps that featurize your data.

..
    Copyright 2019, Neuraxio Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

"""
from neuraxle.pipeline import Pipeline
from neuraxle.steps.flow import ChooseOneOrManyStepsOf
from neuraxle.steps.numpy import NumpyFFT, NumpyAbs, NumpyFlattenDatum, NumpyConcatenateInnerFeatures, NumpyMean, \
    NumpyMedian, NumpyMin, NumpyMax, NumpyArgMax
from neuraxle.union import FeatureUnion


class FFTPeakBinWithValue(FeatureUnion):
    """
    Compute peak fft bins (int), and their magnitudes' value (float), to concatenate them.
    This is intended to be used only after a NumpyFFT absolute step.

    .. seealso::
        :class:`~neuraxle.base.BaseStep`,
        :class:`~neuraxle.base.NonFittableMixin`,
        :class:`~neuraxle.steps.numpy.NumpyFFT`,
        :class:`Cheap3DTo2DTransformer`
    """
    def __init__(self):
        super().__init__([
            NumpyArgMax(axis=-2),
            NumpyMax(axis=-2)
        ], joiner=NumpyConcatenateInnerFeatures())


class Cheap3DTo2DTransformer(ChooseOneOrManyStepsOf):
    """
    Prebuild class to featurize 3D data into 2D data for simple classification or regression, for instance.

    You can enable, or disable features using hyperparams :

    .. code-block:: python

        step = Cheap3DTo2DTransformer().set_hyperparams(hyperparams={
            'FFT__enabled': True,
            'NumpyMean__enabled': True,
            'NumpyMedian__enabled': True,
            'NumpyMin__enabled': True,
            'NumpyMax__enabled': True
        })

    .. seealso::
        :class:`~neuraxle.steps.flow.ChooseOneOrManyStepsOf`,
        :class:`~neuraxle.steps.numpy.NumpyFFT`,
        :class:`~neuraxle.steps.numpy.NumpyAbs`,
        :class:`~neuraxle.steps.numpy.NumpyFlattenDatum`,
        :class:`FFTPeakBinWithValue`,
        :class:`~neuraxle.steps.numpy.NumpyConcatenateInnerFeatures`,
        :class:`~neuraxle.steps.numpy.NumpyMean`,
        :class:`~neuraxle.steps.numpy.NumpyMedian`,
        :class:`~neuraxle.steps.numpy.NumpyMin`,
        :class:`~neuraxle.steps.numpy.NumpyMax`
    """

    def __init__(self):
        super().__init__([
            Pipeline([
                NumpyFFT(),
                NumpyAbs(),
                FeatureUnion([
                    NumpyFlattenDatum(),  # Reshape from 3D to flat 2D: flattening data except on batch size
                    FFTPeakBinWithValue()  # Extract 2D features from the 3D FFT bins
                ], joiner=NumpyConcatenateInnerFeatures())
            ]).set_name('FFT'),
            NumpyMean(),
            NumpyMedian(),
            NumpyMin(),
            NumpyMax()
        ])



"""
Neuraxle's Base Hyperparameter Repository Classes
=================================================

Data objects and related repositories used by AutoML.

Classes are splitted like this for the AutoML:

- Projects
- Clients
- Rounds (runs)
- Trials
- TrialSplits
- MetricResults

..
    Copyright 2022, Neuraxio Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.


"""

import copy
import datetime
import json
import math
import typing
from abc import ABC, abstractmethod
from collections import OrderedDict
from dataclasses import dataclass, field, fields
from json.encoder import JSONEncoder
from typing import (Any, Dict, Generic, List, Optional, Sequence, Tuple, Type,
                    TypeVar, Union)

import numpy as np
from neuraxle.base import CX, BaseService, TrialStatus
from neuraxle.hyperparams.space import HyperparameterSamples, RecursiveDict
from neuraxle.logging.logging import LOGGING_DATETIME_STR_FORMAT


SubDataclassT = TypeVar('SubDataclassT', bound=Optional['BaseDataclass'])
ScopedLocationAttrInt = int
ScopedLocationAttrStr = str
ScopedLocationAttr = Union[str, int]
Slice = Sequence

DEFAULT_PROJECT: ScopedLocationAttrStr = "default_project"
DEFAULT_CLIENT: ScopedLocationAttrStr = "default_client"
DEFAULT_ROUND: ScopedLocationAttrInt = 0
DEFAULT_TRIAL: ScopedLocationAttrInt = 0
DEFAULT_TRIAL_SPLIT: ScopedLocationAttrInt = 0
DEFAULT_METRIC_NAME: ScopedLocationAttrStr = "main"

NULL_PROJECT = "NO_PROJECT"
NULL_CLIENT = "NO_CLIENT"
NULL_ROUND = -math.inf
NULL_TRIAL = -math.inf
NULL_TRIAL_SPLIT = -math.inf
NULL_METRIC_NAME = "NO_METRIC"


def to_int_if_not_float_or_none(attr) -> ScopedLocationAttrInt:
    if isinstance(attr, float) or attr is None:
        return attr
    return int(attr)


def to_str_if_not_none(attr) -> ScopedLocationAttrStr:
    if attr is None or attr == '':
        return None
    return str(attr)


@dataclass(order=True)
class ScopedLocation(BaseService):
    """
    A location in the metadata tree.
    """
    project_name: ScopedLocationAttrStr = None
    client_name: ScopedLocationAttrStr = None
    round_number: ScopedLocationAttrInt = None
    trial_number: ScopedLocationAttrInt = None
    split_number: ScopedLocationAttrInt = None
    metric_name: ScopedLocationAttrStr = None

    def __post_init__(self):
        self.project_name = to_str_if_not_none(self.project_name)
        self.client_name = to_str_if_not_none(self.client_name)
        self.round_number = to_int_if_not_float_or_none(self.round_number)
        self.trial_number = to_int_if_not_float_or_none(self.trial_number)
        self.split_number = to_int_if_not_float_or_none(self.split_number)
        self.metric_name = to_str_if_not_none(self.metric_name)
        BaseService.__init__(self)

    # get the field value from BaseMetadata subclass:
    def __getitem__(
        self, key: Union[Type['BaseDataclass'], Slice[Type['BaseDataclass']]]
    ) -> Union[ScopedLocationAttr, 'ScopedLocation']:
        """
        Get sublocation attr from the provided :class:`BaseMetadata` type,
        or otherwise get a slice of the same type, sliced from a :class:`BaseMetadata` type range of attributes to keep.
        If the key is a slice, the end loc is kept as inclusive.
        """
        if key is None:
            return ScopedLocation()

        if isinstance(key, slice):
            if key.stop is None or key.start is not None:
                raise ValueError("Slice stop must be specified and slice start must be None.")

            idx: int = key.stop
            if not isinstance(idx, int):
                if idx == RootDataclass:
                    return ScopedLocation()
                idx: int = list(dataclass_2_id_attr.keys()).index(key.stop) + 1

            return ScopedLocation(
                *self.as_list()[:idx]
            )

        if key == RootDataclass:
            return None

        if isinstance(key, int):
            if key < 0:
                if len(self) == 0:
                    return None
                else:
                    return self.as_list()[key]
            key = list(dataclass_2_id_attr.keys())[key]

        if key in dataclass_2_id_attr.keys():
            return getattr(self, dataclass_2_id_attr[key])

        raise ValueError(f"Invalid key type {key.__class__.__name__} for key {key}.")

    def _copy(self) -> 'ScopedLocation':
        """
        Returns a copy of the :class:`ScopedLocation`.
        """
        return ScopedLocation(*self.as_list())

    def with_dc(self, dc: 'BaseDataclass') -> 'ScopedLocation':
        """
        Returns a new :class:`ScopedLocation` with the provided :class:`BaseDataclass` (dc) type's id added.
        """
        if isinstance(dc, RootDataclass):
            return ScopedLocation()
        elif dc.is_terminal_leaf():
            cpy = self._copy()
            cpy.metric_name = dc.get_id()
            return cpy
        self_copy = self._copy()
        self_copy[dc.__class__] = dc.get_id()
        return self_copy

    def fill_to_dc(self, dc: 'BaseDataclass') -> 'ScopedLocation':
        """
        Returns a :class:`ScopedLocation` with the provided :class:`BaseDataclass` (dc) type's id added at the end, with the particularity that if some elements are missing, they are filled with the default null values.
        """
        expected_len = list(dataclass_2_subdataclass.keys()).index(dc.__class__)
        _len = len(self)
        # if the length is not the expected one, fill with None attrs (for the missing ones). Otherwise, return the current scoped location reduced at the good dc depth.
        if _len > expected_len:
            return self.at_dc(dc)
        else:
            return self.pad_nans().at_dc(dc).popped().with_dc(dc)

    def pad_nans(self) -> 'ScopedLocation':
        """
        Returns a :class:`ScopedLocation` with the missing elements filled with the default null values.
        """
        self_copy = self._copy()
        null_vals = [
            NULL_PROJECT,
            NULL_CLIENT,
            NULL_ROUND,
            NULL_TRIAL,
            NULL_TRIAL_SPLIT,
            NULL_METRIC_NAME,
        ]
        for i in range(len(self_copy), len(dataclass_2_id_attr)):
            self_copy[list(dataclass_2_id_attr.keys())[i]] = null_vals[i]
        return self_copy

    def with_id(self, _id: ScopedLocationAttr) -> 'ScopedLocation':
        """
        Returns a longer :class:`ScopedLocation` with the provided id added at the end.
        """
        return ScopedLocation(*(self.as_list() + [_id]))

    def at_dc(self, dc: 'BaseDataclass') -> 'ScopedLocation':
        """
        Returns a trimmed :class:`ScopedLocation` with the provided :class:`BaseDataclass` (dc) type's id as the ScopedLocation's deepest attribute.
        """
        if isinstance(dc, RootDataclass):
            return ScopedLocation()
        return self[:dc.__class__]

    @staticmethod
    def default_full() -> 'ScopedLocation':
        """
        Returns a :class:`ScopedLocation` with all attributes
        set to the default non-null value instead of None.
        """
        return ScopedLocation(
            DEFAULT_PROJECT, DEFAULT_CLIENT, DEFAULT_ROUND, DEFAULT_TRIAL, DEFAULT_TRIAL_SPLIT, DEFAULT_METRIC_NAME)

    @staticmethod
    def default(
        round_number: Optional[ScopedLocationAttrInt] = None,
        trial_number: Optional[ScopedLocationAttrInt] = None,
        split_number: Optional[ScopedLocationAttrInt] = None,
        metric_name: Optional[ScopedLocationAttrStr] = None,
    ) -> 'ScopedLocation':
        """
        Returns the default :class:`ScopedLocation`. That is:

        .. code-block:: python
            ScopedLocation("default_project", "default_client", *kargs, **kwargs).

        """
        if not ((round_number is not None) >= (trial_number is not None) >= (split_number is not None) >= (metric_name is not None)):
            raise ValueError(
                "round_number, trial_number, split_number, and metric_name "
                "must be specified in order, one after the other.")

        _args = []
        if round_number is not None:
            _args.append(round_number)
            if trial_number is not None:
                _args.append(trial_number)
                if split_number is not None:
                    _args.append(split_number)
                    if metric_name is not None:
                        _args.append(metric_name)

        return ScopedLocation(DEFAULT_PROJECT, DEFAULT_CLIENT, *_args)

    def __setitem__(self, key: Type['BaseDataclass'], value: ScopedLocationAttr):
        """
        Set sublocation attr from the provided :class:`BaseMetadata` type.
        """
        curr_attr_to_set_idx: int = len(self)
        key_idx: int = list(dataclass_2_id_attr.keys()).index(key)

        if curr_attr_to_set_idx == key_idx + 1 and self[key] == value:
            # operation is redundant but ok: no effect.
            return
        elif curr_attr_to_set_idx == key_idx:
            # update as expected:
            key_attr_name: str = dataclass_2_id_attr[key]
            setattr(self, key_attr_name, value)
        else:
            # Throw value error if the key's type is not yet to be defined:
            # curr_attr_to_set_idx != key_idx:
            raise ValueError(
                f"{key} is not yet to be defined into {self}.")

    def peek(self) -> ScopedLocationAttr:
        """
        Pop without removing the last element: return the last non-None element.
        """
        return self.as_list()[-1]

    def pop(self) -> ScopedLocationAttr:
        """
        Returns the last `not None` scoped location attribute and remove it from self.
        """
        for attr_name in reversed(dataclass_2_id_attr.values()):
            attr_value = getattr(self, attr_name)
            if attr_value is not None:
                setattr(self, attr_name, None)
                return attr_value

    def popped(self) -> 'ScopedLocation':
        """
        Returns a new :class:`ScopedLocation` with the last `not None` scoped location attribute removed.
        """
        return ScopedLocation(*self.as_list()[:-1])

    def __len__(self) -> int:
        """
        Returns the number of not none scoped location attributes.
        """
        return len(self.as_list())

    def as_list(self, stringify: bool = False) -> List[ScopedLocationAttr]:
        """
        Returns a list of the scoped location attributes.
        Item that has a value of None are not included in the list.

        :param stringify: If True, the scoped location attributes are converted to strings.
        :return: list of not none scoped location attributes
        """
        _list: List[ScopedLocationAttr] = [
            getattr(self, attr_name)
            for attr_name in dataclass_2_id_attr.values()
        ]
        _list_ret = []
        for i in _list + [None]:
            if i is not None:
                if stringify:
                    i = str(i)
                _list_ret.append(i)
            else:
                break
        return _list_ret

    def new_dataclass_from_id(self):
        """
        Creates a new :class:`BaseDataclass` of the right type with
        just the provided ID filled.
        """
        if len(self) == 0:
            return RootDataclass()
        dataklass = self.last_dc_type()

        return dataklass().set_id(self.as_list()[-1])

    def last_dc_type(self) -> Type['BaseDataclass']:
        if len(self) == 0:
            return RootDataclass
        dataklass: Type[BaseDataclass] = list(dataclass_2_id_attr.keys())[len(self) - 1]
        return dataklass

    def __eq__(self, other: 'ScopedLocation') -> bool:
        return self.as_list() == other.as_list()

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({repr(self.as_list())[1:-1]})"

    def __str__(self) -> str:
        return self.__repr__()


@dataclass(order=True)
class BaseDataclass(Generic[SubDataclassT], ABC):

    def __post_init__(self):
        self._validate()

    def _validate(self):
        for k, v in zip(self.get_sublocation_keys(), self.get_sublocation_values()):
            if isinstance(v, dict) and '__type__' in v:
                assert v['__type__'] == dataclass_2_subdataclass[self.__class__].__name__
                v: SubDataclassT = BaseDataclass.from_dict(v)
                assert v.get_id() == k
                self.store(v)

        subdataklass: Type[SubDataclassT] = dataclass_2_subdataclass[self.__class__]
        if subdataklass is not None and not all(
            (isinstance(s, subdataklass) for s in self.get_sublocation_values() if s is not None)
        ):
            raise ValueError(f"{self.__class__.__name__} must have all sublocation values as {subdataklass.__name__}.")

    @staticmethod
    def from_dict(_dict: Dict[str, Any]) -> 'BaseDataclass':
        if '__type__' not in _dict:
            raise ValueError("Dict must have a __type__ key.")
        _dict: dict = copy.copy(_dict)
        _type_str: str = _dict.pop('__type__')
        _klass: type = str_2_dataclass[_type_str]

        return _klass(**_dict)

    def to_dict(self) -> Dict[str, Any]:
        _dict = OrderedDict()
        _dict['__type__'] = self.__class__.__name__
        for _field in fields(self):
            _attr = getattr(self, _field.name)
            if _attr is not None:
                if isinstance(_attr, BaseDataclass):
                    _attr = _attr.to_dict()
                _dict[_field.name] = _attr
        return _dict

    @property
    def _id_attr_name(self) -> str:
        return dataclass_2_id_attr[self.__class__]

    @property
    def _sublocation_attr_name(self) -> str:
        return dataclass_2_subloc_attr[self.__class__]

    def get_id(self) -> ScopedLocationAttr:
        return getattr(self, self._id_attr_name)

    def set_id(self, _id: ScopedLocationAttr) -> 'BaseDataclass':
        setattr(self, self._id_attr_name, _id)
        self._validate()
        return self

    def get_sublocation(self) -> Union[List[SubDataclassT], 'OrderedDict[str, SubDataclassT]']:
        return getattr(self, dataclass_2_subloc_attr[self.__class__])

    def get_sublocation_items(self) -> List[Tuple[ScopedLocationAttr, SubDataclassT]]:
        return [(k, v) for k, v in zip(self.get_sublocation_keys(), self.get_sublocation_values())]

    def set_sublocation(self, sublocation: Union[List[SubDataclassT], 'OrderedDict[str, SubDataclassT]']) -> 'BaseDataclass':
        setattr(self, dataclass_2_subloc_attr[self.__class__], sublocation)
        return self

    @abstractmethod
    def set_sublocation_keys(self, keys: List[ScopedLocationAttr]) -> 'BaseDataclass':
        """
        Use this to set a shallow sublocation only from their keys.
        """
        raise NotImplementedError("This is an abstract method.")

    def __contains__(self, key: ScopedLocation) -> bool:
        # Too shallow:
        if len(key) < list(dataclass_2_subdataclass.keys()).index(self.__class__):
            raise ValueError(f"Key not deep enough for this dataclass of type {self.__class__.__name__}")

        # Terminal recursion condition:
        if key.at_dc(self) == key:
            return key.peek() == self.get_id()

        # Recursive call deeper otherwise:
        key_subattr: ScopedLocationAttr = key[self.subdataclass_type()]
        if self._is_key_in_subattr(key_subattr):
            return key in self.get_sublocation()[key_subattr]
        return False

    def __getitem__(self, loc: ScopedLocation) -> 'SubDataclassT':
        subdataklass: Type[SubDataclassT] = dataclass_2_subdataclass[self.__class__]
        if subdataklass is None:
            return self
        located_sub_attr: ScopedLocationAttr = loc[subdataklass]
        if located_sub_attr is None:
            return self
        if not self._is_key_in_subattr(located_sub_attr):
            raise KeyError(f"Item at loc={loc} is not in {self.shallow()}.")
        subloc: SubDataclassT = self.get_sublocation()[located_sub_attr]
        sublocs_sublocs: SubDataclassT = subloc[loc]
        return sublocs_sublocs

    def _is_key_in_subattr(self, key_subattr):
        return key_subattr in self.get_sublocation_keys()

    def __len__(self) -> int:
        return len(self.get_sublocation())

    @abstractmethod
    def store(self, dc: SubDataclassT) -> ScopedLocationAttr:
        """
        Add a subdataclass to the sublocation, at its proper ID.
        """
        raise NotImplementedError("Must use mixins.")

    @abstractmethod
    def shallow(self) -> 'BaseDataclass':
        """
        Replaces the sublocation items with None when the sublocation is a BaseDataclass type.
        """
        raise NotImplementedError("Must use mixins.")

    @abstractmethod
    def empty(self) -> 'BaseDataclass':
        """
        Do empty the sublocation when the sublocation is a BaseDataclass type.
        """
        raise NotImplementedError("Must use mixins.")

    def has_sublocation_dataclasses(self) -> bool:
        return self.__class__ in list(dataclass_2_subdataclass.keys())[:-1]

    @abstractmethod
    def get_sublocation_values(self) -> List[SubDataclassT]:
        raise NotImplementedError("Must use mixins.")

    @abstractmethod
    def get_sublocation_keys(self) -> List[ScopedLocationAttr]:
        raise NotImplementedError("Must use mixins.")

    @classmethod
    def subdataclass_type(cls) -> Type[SubDataclassT]:
        return dataclass_2_subdataclass[cls]

    def is_terminal_leaf(self) -> bool:
        return self.__class__ == MetricResultsDataclass

    def _tree(self, _list: List[ScopedLocation], parent_scope: ScopedLocation) -> List[ScopedLocation]:
        this_scope = parent_scope.fill_to_dc(self)
        _list.append(this_scope)
        if not self.is_terminal_leaf():
            for _subloc in self.get_sublocation_values():
                _subloc._tree(_list, this_scope)
        return _list

    def tree(self) -> List[ScopedLocation]:
        return self._tree([], ScopedLocation())


@dataclass(order=True)
class DataclassHasOrderedDictMixin:

    def _validate(self):
        self._sort()
        super()._validate()
        if not isinstance(self.get_sublocation(), OrderedDict):
            raise ValueError(f"{self.__class__.__name__} must have an OrderedDict as sublocation.")
        if not all(
            (isinstance(s, ScopedLocationAttrStr) for s in self.get_sublocation_keys())
        ):
            raise ValueError(f"{self.__class__.__name__} must have all sublocation keys as strings.")

    def get_sublocation(self) -> 'OrderedDict[ScopedLocationAttrStr, SubDataclassT]':
        ret = super().get_sublocation()
        return ret

    def set_sublocation(self, sublocation: 'OrderedDict[ScopedLocationAttrStr, SubDataclassT]') -> None:
        setattr(self, self._sublocation_attr_name, sublocation)
        self._validate()

    def set_sublocation_keys(self, keys: List[ScopedLocationAttr]) -> 'BaseDataclass':
        # TODO: when using a JSON repo, loaded keys are simply sorted alphabetically. That may be a problem for the main metric and metrics' sorting in the future.
        _sublocation = OrderedDict([(str(k), None) for k in keys])
        setattr(self, self._sublocation_attr_name, _sublocation)
        self._validate()

    def set_sublocation_items(
        self, items: List[Tuple[ScopedLocationAttr, SubDataclassT]]
    ) -> 'BaseDataclass':
        _sublocation = OrderedDict([(str(k), v) for k, v in items])
        setattr(self, self._sublocation_attr_name, _sublocation)
        self._validate()
        return self

    def get_sublocation_values(self) -> List[SubDataclassT]:
        return list(self.get_sublocation().values())

    def get_sublocation_keys(self) -> List[ScopedLocationAttrStr]:
        return list(self.get_sublocation().keys())

    def store(self, dc: SubDataclassT) -> ScopedLocationAttrStr:
        self.get_sublocation()[dc.get_id()] = dc
        self._validate()
        return dc.get_id()

    def shallow(self) -> 'BaseDataclass':
        self_copy = copy.copy(self)
        new_sublocation = OrderedDict([(k, None) for k in self.get_sublocation().keys()])
        self_copy.set_sublocation(new_sublocation)
        # Deep copy only after trimming sublocations.
        return copy.deepcopy(self_copy)

    def empty(self) -> 'BaseDataclass':
        self_copy = copy.copy(self)
        self_copy.set_sublocation(OrderedDict())
        return copy.deepcopy(self_copy)

    def _sort(self):
        _sublocation = self.get_sublocation()
        _sublocation = OrderedDict(list(sorted(_sublocation.items())))
        setattr(self, self._sublocation_attr_name, _sublocation)


@dataclass(order=True)
class DataclassHasListMixin:

    def _validate(self):
        super()._validate()
        if not isinstance(self.get_sublocation(), List):
            raise ValueError(f"{self.__class__.__name__} must have a List as sublocation.")

    def get_sublocation(self) -> List[SubDataclassT]:
        return super().get_sublocation()

    def set_sublocation(self, sublocation: List[SubDataclassT]) -> None:
        setattr(self, self._sublocation_attr_name, sublocation)
        self._validate()

    def set_sublocation_keys(self, keys: List[ScopedLocationAttr]) -> 'BaseDataclass':
        _sublocation = [None for k in keys]
        # assert (set(range(len(_sublocation))) == set([int(k) for k in keys])), (
        #     f"Bad sublocation keys are being set into DataclassHasListMixin "
        #     f"(type {self.__class__.__name__}, id={self.get_id()}) : {keys}.")
        setattr(self, self._sublocation_attr_name, _sublocation)
        self._validate()

    def set_sublocation_items(
        self, items: List[Tuple[ScopedLocationAttr, SubDataclassT]]
    ) -> 'BaseDataclass':
        _sublocation = []
        items = [(int(k), v) for k, v in items]
        for i, (k, v) in enumerate(sorted(items)):
            if i != k:
                raise ValueError(
                    f"Bad sublocation keys are being set into DataclassHasListMixin (type {self.__class__.__name__}, id={self.get_id()}) : {items}.")
            assert k == v.get_id() and (i == k), (
                f"Bad sublocation keys are being set into DataclassHasListMixin of type {self.__class__.__name__}, id={self.get_id()} : {i} != {k} != {v.get_id()}. \n\nFor more context: {items}"
            )
            _sublocation.append(v)
        setattr(self, self._sublocation_attr_name, _sublocation)
        self._validate()
        return self

    def get_sublocation_values(self) -> List[SubDataclassT]:
        return self.get_sublocation()

    def get_sublocation_keys(self) -> List[ScopedLocationAttrInt]:
        return list(range(len(self)))

    def _is_key_in_subattr(self, key_subattr):
        return key_subattr in self.get_sublocation_keys() or (key_subattr == -1 and len(self) != 0)

    def store(self, dc: SubDataclassT) -> ScopedLocationAttrInt:
        _id = dc.get_id()
        if _id == self.get_next_i():
            self.get_sublocation().append(dc)
        elif _id < self.get_next_i():
            self.get_sublocation()[dc.get_id()] = dc
        else:
            raise ValueError(
                f"{dc.shallow()} has id {dc.get_id()} which is greater than the next id {self.get_next_i()}.")
        self._validate()
        return _id

    def shallow(self) -> 'BaseDataclass':
        self_copy = copy.copy(self)
        new_sublocation = [None for _ in range(len(self))]
        self_copy.set_sublocation(new_sublocation)
        # Deep copy only after trimming sublocations.
        return copy.deepcopy(self_copy)

    def empty(self) -> 'BaseDataclass':
        self_copy = copy.copy(self)
        self_copy.set_sublocation([])
        return copy.deepcopy(self_copy)

    def get_next_i(self) -> ScopedLocationAttrInt:
        return len(self.get_sublocation())


@dataclass(order=True)
class BaseTrialDataclassMixin:
    """
    Mixin class for :class:`TrialMetadata` and :class:`TrialSplitMetadata` that
    also must inherit from :class:`BaseMetadata`.
    """
    hyperparams: HyperparameterSamples = field(default_factory=HyperparameterSamples)
    status: TrialStatus = TrialStatus.PLANNED
    created_time: datetime.datetime = field(default_factory=datetime.datetime.now)
    start_time: datetime.datetime = None
    end_time: datetime.datetime = None
    # error: str = None
    # error_traceback: str = None
    # logs: List[str] = field(default_factory=list)

    def __post_init__(self):
        self.hyperparams = HyperparameterSamples(self.hyperparams)
        BaseDataclass.__post_init__(self)

    def _validate(self):
        super()._validate()
        if not isinstance(self.hyperparams, HyperparameterSamples):
            self.hyperparams = HyperparameterSamples(self.hyperparams)
        self._validate_date_attr("created_time")
        self._validate_date_attr("start_time")
        self._validate_date_attr("end_time")
        if not isinstance(self.status, TrialStatus):
            self.status = TrialStatus(self.status)

    def _validate_date_attr(self, attr_name) -> None:
        attr_value = getattr(self, attr_name)
        if attr_value is not None and not isinstance(attr_value, datetime.datetime):
            attr_value = datetime.datetime.strptime(attr_value, LOGGING_DATETIME_STR_FORMAT)
            setattr(self, attr_name, attr_value)

    def start(self) -> 'BaseTrialDataclassMixin':
        self.status = TrialStatus.RUNNING
        self.start_time = datetime.datetime.now()
        self._validate()
        return self

    def end(self, status: TrialStatus) -> 'BaseTrialDataclassMixin':
        self.status = status
        self.end_time = datetime.datetime.now()
        if self.start_time is None:
            self.start_time = self.created_time
        self._validate()
        return self


@dataclass(order=True)
class RootDataclass(DataclassHasOrderedDictMixin, BaseDataclass['ProjectDataclass']):
    projects: typing.OrderedDict[str, 'ProjectDataclass'] = field(
        default_factory=lambda: OrderedDict({DEFAULT_PROJECT: ProjectDataclass()}))

    @property
    def _id_attr_name(self) -> str:
        return None

    @property
    def _sublocation_attr_name(self) -> str:
        return "projects"

    def get_id(self) -> ScopedLocationAttr:
        return None

    def get_sublocation(self) -> 'OrderedDict[str, SubDataclassT]':
        return self.projects

    # def _validate(self):
    #     super()._validate()
    #     if DEFAULT_PROJECT not in self.get_sublocation_keys():
    #         raise ValueError(f"{DEFAULT_PROJECT} should be in {self.__class__.__name__}. Got {self}.")


@dataclass(order=True)
class ProjectDataclass(DataclassHasOrderedDictMixin, BaseDataclass['ClientDataclass']):
    project_name: str = DEFAULT_PROJECT
    clients: typing.OrderedDict[str, 'ClientDataclass'] = field(
        default_factory=lambda: OrderedDict({DEFAULT_CLIENT: ClientDataclass()}))

    # def _validate(self):
    #     super()._validate()
    #     if DEFAULT_CLIENT not in self.get_sublocation_keys():
    #         raise ValueError(f"{DEFAULT_CLIENT} should be in {self.__class__.__name__}. Got {self}.")


@dataclass(order=True)
class ClientDataclass(DataclassHasListMixin, BaseDataclass['RoundDataclass']):
    client_name: ScopedLocationAttrStr = DEFAULT_CLIENT
    rounds: List['RoundDataclass'] = field(default_factory=list)


@dataclass(order=True)
class RoundDataclass(DataclassHasListMixin, BaseDataclass['TrialDataclass']):
    round_number: ScopedLocationAttrInt = 0
    trials: List['TrialDataclass'] = field(default_factory=list)
    main_metric_name: str = None


@dataclass(order=True)
class TrialDataclass(DataclassHasListMixin, BaseTrialDataclassMixin, BaseDataclass['TrialSplitDataclass']):
    """
    This class is a data structure most often used under :class:`AutoML` to store information about a trial.
    This information is itself managed by the :class:`HyperparameterRepository` class
    and the :class:`Trial` class within the AutoML.
    """
    trial_number: ScopedLocationAttrInt = 0
    validation_splits: List['TrialSplitDataclass'] = field(default_factory=list)
    retrained_split: 'TrialSplitDataclass' = None

    def store(self, dc: SubDataclassT) -> ScopedLocationAttrInt:
        if dc.get_id() == RETRAIN_TRIAL_SPLIT_ID:
            self.retrained_split = dc
            return RETRAIN_TRIAL_SPLIT_ID
        else:
            return super().store(dc)

    def __contains__(self, key: ScopedLocation) -> bool:
        if key[self.__class__] != self.get_id():
            return False
        if key[self.subdataclass_type()] == RETRAIN_TRIAL_SPLIT_ID:
            if self.retrained_split is None:
                return False
            return key in self.retrained_split
        else:
            return super().__contains__(key)

    def __getitem__(self, loc: ScopedLocation) -> SubDataclassT:
        if loc[self.__class__] != self.get_id():
            raise KeyError(f"Item at loc={loc} is not in {self.shallow()}.")
        if loc[self.subdataclass_type()] == RETRAIN_TRIAL_SPLIT_ID:
            if self.retrained_split is None:
                raise KeyError(f"Item at loc={loc} is not in {self.shallow()}. No retrains are available.")
            return self.retrained_split[loc]
        else:
            return super().__getitem__(loc)

    def set_sublocation_keys(self, keys: List[ScopedLocationAttr]) -> 'BaseDataclass':
        keys = [int(k) for k in keys]
        if RETRAIN_TRIAL_SPLIT_ID in keys:
            # TODO: should getters take into account RETRAIN_TRIAL_SPLIT_ID as well?
            self._set_shallow_retrained_split_id()
            keys.remove(RETRAIN_TRIAL_SPLIT_ID)
        super().set_sublocation_keys(keys)

    def set_sublocation_items(self, items: List[Tuple[ScopedLocationAttr, SubDataclassT]]) -> 'BaseDataclass':
        items = [(int(k), v) for k, v in items]
        items_filtered = []
        for item in items:
            if item[0] == RETRAIN_TRIAL_SPLIT_ID:
                self.retrained_split = item[1]
            else:
                items_filtered.append(item)
        super().set_sublocation_items(items_filtered)

    def shallow(self) -> 'BaseDataclass':
        shallowed: TrialDataclass = super().shallow()
        if shallowed.retrained_split is not None and not isinstance(shallowed.retrained_split, str):
            assert shallowed.retrained_split.get_id() == RETRAIN_TRIAL_SPLIT_ID, (
                f"Retrain split has a id != -1: {shallowed.retrained_split}")
            shallowed._set_shallow_retrained_split_id()
        return shallowed

    def _set_shallow_retrained_split_id(self):
        self.retrained_split = "ShallowedRetrainedSplitWithId(-1)"

    def empty(self) -> 'BaseDataclass':
        emptied: TrialDataclass = super().empty()
        emptied.retrained_split = None
        return emptied


RETRAIN_TRIAL_SPLIT_ID = -1


@dataclass(order=True)
class TrialSplitDataclass(DataclassHasOrderedDictMixin, BaseTrialDataclassMixin, BaseDataclass['MetricResultsDataclass']):
    """
    TrialSplit object used by AutoML algorithm classes.
    """
    split_number: ScopedLocationAttrInt = 0
    metric_results: typing.OrderedDict[str, 'MetricResultsDataclass'] = field(default_factory=OrderedDict)
    # introspection_data: RecursiveDict[str, Number] = field(default_factory=RecursiveDict)

    def is_retrain_split(self) -> bool:
        return self.split_number == -1


@dataclass(order=True)
class MetricResultsDataclass(DataclassHasListMixin, BaseDataclass[float]):
    """
    MetricResult object used by AutoML algorithm classes.
    """
    metric_name: ScopedLocationAttrStr = DEFAULT_METRIC_NAME
    validation_values: List[float] = field(default_factory=list)  # one per epoch.
    train_values: List[float] = field(default_factory=list)
    higher_score_is_better: bool = True

    def shallow(self) -> 'BaseDataclass':
        return copy.deepcopy(self)

    def empty(self) -> 'BaseDataclass':
        return copy.deepcopy(self)


dataclass_2_id_attr: typing.OrderedDict[BaseDataclass, str] = OrderedDict([
    (ProjectDataclass, "project_name"),
    (ClientDataclass, "client_name"),
    (RoundDataclass, "round_number"),
    (TrialDataclass, "trial_number"),
    (TrialSplitDataclass, "split_number"),
    (MetricResultsDataclass, "metric_name"),
])

dataclass_2_subloc_attr: typing.OrderedDict[BaseDataclass, str] = OrderedDict([
    (RootDataclass, "projects"),
    (ProjectDataclass, "clients"),
    (ClientDataclass, "rounds"),
    (RoundDataclass, "trials"),
    (TrialDataclass, "validation_splits"),
    (TrialSplitDataclass, "metric_results"),
    (MetricResultsDataclass, "validation_values"),
])

dataclass_2_subdataclass: typing.OrderedDict[BaseDataclass, SubDataclassT] = OrderedDict([
    (RootDataclass, ProjectDataclass),
    (ProjectDataclass, ClientDataclass),
    (ClientDataclass, RoundDataclass),
    (RoundDataclass, TrialDataclass),
    (TrialDataclass, TrialSplitDataclass),
    (TrialSplitDataclass, MetricResultsDataclass),
    (MetricResultsDataclass, None),
])

str_2_dataclass: typing.OrderedDict[str, BaseDataclass] = OrderedDict([
    (RootDataclass.__name__, RootDataclass),
    (ProjectDataclass.__name__, ProjectDataclass),
    (ClientDataclass.__name__, ClientDataclass),
    (RoundDataclass.__name__, RoundDataclass),
    (TrialDataclass.__name__, TrialDataclass),
    (TrialSplitDataclass.__name__, TrialSplitDataclass),
    (MetricResultsDataclass.__name__, MetricResultsDataclass),
    (RecursiveDict.__name__, RecursiveDict),
])


def as_named_odict(
    obj: Union[BaseDataclass, List[BaseDataclass]]
) -> 'OrderedDict[ScopedLocationAttrStr, BaseDataclass]':

    if isinstance(obj, BaseDataclass):
        obj = [obj]

    return OrderedDict([
        (i.get_id(), i) for i in obj
    ])


def object_pairs_decoder(obj):
    return object_decoder(obj, odictify=True)


def object_decoder(obj, odictify=False):
    if odictify:
        obj = OrderedDict(obj)
    if '__type__' in obj and obj['__type__'] in str_2_dataclass:
        # cls: Type = str_2_dataclass[obj['__type__']]
        return BaseDataclass.from_dict(obj)
        # kwargs = dict(obj)
        # del kwargs['__type__']
        # return cls(**kwargs)
    return obj


class MetadataJSONEncoder(JSONEncoder):
    def default(self, o):
        if isinstance(o, np.ndarray):
            return o.tolist()
        if isinstance(o, TrialStatus):
            return o.value
        if isinstance(o, RecursiveDict):
            return o.to_flat_dict()
        if isinstance(o, datetime.datetime):
            return o.strftime(LOGGING_DATETIME_STR_FORMAT)
        if isinstance(o, BaseDataclass):
            return o.to_dict()
        return JSONEncoder.encode(self, o)


def to_json(obj: BaseDataclass) -> str:
    return json.dumps(obj, cls=MetadataJSONEncoder)


def from_json(_json: str) -> BaseDataclass:
    return json.loads(_json, object_pairs_hook=object_pairs_decoder, object_hook=object_decoder)



"""
Streaming Pipelines for Parallel and Queued Data Processing
===================================================================

Neuraxle steps for streaming data in parallel in the pipeline.

Pipelines can stream data in queues with workers for each steps.
Max queue sizes can be set, as well as number of clones per steps
for the transformers.


..
    Copyright 2022, Neuraxio Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

"""
import pickle
import queue
import traceback
from abc import abstractmethod
from collections import OrderedDict, defaultdict
from multiprocessing import Lock, Process, Queue, RLock
from multiprocessing.dummy import current_process
from threading import Thread, current_thread
from typing import Any, Dict, Iterable, List, Optional, Tuple, Union

from neuraxle.base import BaseSaver, BaseTransformer
from neuraxle.base import ExecutionContext as CX
from neuraxle.base import MetaStep, MixinForBaseTransformer, NamedStepsList, NonFittableMixin, _FittableStep
from neuraxle.data_container import DACT, DIT, EOT, IDT, ListDataContainer, PredsDACT, StripAbsentValues
from neuraxle.hyperparams.space import RecursiveDict
from neuraxle.logging.logging import (ParallelLoggingConsumerThread,
                                      register_log_producer_for_main_logger_thread_to_consume)
from neuraxle.pipeline import Joiner, MiniBatchSequentialPipeline, Pipeline
from neuraxle.steps.numpy import NumpyConcatenateOuterBatch


class _ProducerConsumerStepSaver(BaseSaver):
    """
    Saver for :class:`_ProducerConsumerMixin`.
    This saver class makes sure that the non-picklable queue
    is deleted upon saving for multiprocessing steps.
    """

    def save_step(self, step: BaseTransformer, context: 'CX') -> BaseTransformer:
        step: _ProducerConsumerMixin = step  # typing.
        step._allow_exit_without_queue_flush()
        step.input_queue = None
        step.consumers = []
        return step

    def can_load(self, step: BaseTransformer, context: 'CX') -> bool:
        return True

    def load_step(self, step: 'BaseTransformer', context: 'CX') -> 'BaseTransformer':
        step: _ProducerConsumerMixin = step  # typing.
        step.input_queue = None
        return step


class _QueueDestroyedError(EOFError):
    """
    Error raised when the queue is destroyed.
    """
    pass


EMPTY_CONSUMER_QUEUE_TIMEOUT_SECS = 0.1


class _ProducerConsumerMixin(MixinForBaseTransformer):
    """
    A class to represent a step that can receive minibatches from a producer in its queue to consume them.
    Once minibatches are consumed by the present step, they are produced back to the next consumers in line.

    The Queue in self is the one at the entry to be consumed by self. The output queue is external, in the registred consumers.

    Therefore, in this sense, consumers and producers are both themselves instances of _HasQueueMixin and play the two roles unless they are the first an lasts of their multiprocessing pipelines.
    They will thus be used to consume the produced tasks added to them in their other threads or processes.
    """

    def __init__(self, max_queued_minibatches: int = 0):
        # TODO: two class for this: a producer class and a consumer class perhaps?
        MixinForBaseTransformer.__init__(self)
        self.savers.append(_ProducerConsumerStepSaver())

        self.max_queued_minibatches: int = max(0, max_queued_minibatches or 0)

        self.input_queue: Queue = None
        self.consumers: List[Queue] = []

    def _setup(self, context: 'CX' = None) -> Optional[RecursiveDict]:
        self._init_queue()

    def _init_queue(self):
        if self.input_queue is None:
            self.input_queue = Queue(maxsize=self.max_queued_minibatches)

    def _teardown(self) -> Optional[RecursiveDict]:
        self.join()
        self.input_queue = None
        return RecursiveDict()

    def register_consumer(self, recepient: '_ProducerConsumerMixin') -> '_ProducerConsumerMixin':
        """
        Add a consumer to self.consumers so that when self produces a minibatch, it allows consumers to consume it.
        """
        self._init_queue()
        recepient._init_queue()
        self.consumers.append(recepient.input_queue)
        return self

    def put_minibatch_produced_to_next_consumers(self, task: 'QueuedMinibatchTask'):
        """
        Push a minibatch to all subsequent consumers to allow them to consume it. If the task is terminal, close our own queue.
        """
        task.worker_name = self.name
        self._ensure_task_picklable(task)
        for consumers_of_self in self.consumers:
            consumers_of_self: _ProducerConsumerMixin  # typing...
            consumers_of_self.put(task)

    def put_minibatch_produced(self, task: 'QueuedMinibatchTask'):
        """
        Put a minibatch in queue.
        The caller of this method is the producer of the minibatch and is external to self.
        """
        self._ensure_task_picklable(task)
        try:
            self.input_queue.put(task)
            if task.is_error():
                self._allow_exit_without_queue_flush()
        except Exception as e:
            raise _QueueDestroyedError("It seems like the queue to produce to has been destroyed.") from e

    def _ensure_task_picklable(self, task: 'QueuedMinibatchTask'):
        try:
            pickle.dumps(task)
        except Exception as e:
            # See: https://github.com/python/cpython/issues/79423
            raise pickle.PickleError(f"Couldn't pickle task {task} to send it to the next consumers.") from e

    def _get_minibatch_to_consume(self) -> 'QueuedMinibatchTask':
        """
        Get last minibatch in queue. The caller of this method is probably self,
        that is why the method is private (starts with an underscore).
        This method can raise an EOFError.
        """
        try:
            task: QueuedMinibatchTask = self.input_queue.get(block=True, timeout=EMPTY_CONSUMER_QUEUE_TIMEOUT_SECS)
            if task.is_error():
                self._allow_exit_without_queue_flush()
            return task
        except queue.Empty as e:
            # only happens when block=False or with a timeout:
            raise e from e
        except Exception as e:
            raise _QueueDestroyedError("It seems like the queue to consume from has been destroyed.") from e

    def join(self):
        if self.input_queue is not None:
            self.input_queue.close()
            self._allow_exit_without_queue_flush()
            self.input_queue.join_thread()
            self.input_queue = None

    def _allow_exit_without_queue_flush(self):
        if self.input_queue is not None:
            self.input_queue.cancel_join_thread()


class QueuedMinibatchTask:
    """
    Data object to contain the minibatch processed by producers and consumers.
    """

    def __init__(self, minibatch_dact: DACT, step_name: str = None):
        self.worker_name: str = step_name
        self.minibatch_dact: ListDataContainer = minibatch_dact

    def is_error(self) -> bool:
        return False

    def to_error(self, error: Exception) -> 'MinibatchError':
        _tb_str: str = traceback.format_exc()
        _thread_name: str = current_thread().name
        _process_name: str = current_process().name
        traceback_msg = f'{error}\n\nProcess:Thread "{_process_name}:{_thread_name}" {_tb_str}'

        return MinibatchError(
            minibatch_dact=self.minibatch_dact,
            step_name=self.worker_name,
            err=error,
            traceback_msg=traceback_msg)


class MinibatchError(QueuedMinibatchTask):
    """
    Data object to represent an error in a minibatch.
    """

    def __init__(self, minibatch_dact: DACT, step_name: str, err: Exception, traceback_msg=None):
        super().__init__(minibatch_dact, step_name)
        self.err: Exception = err
        self.traceback_msg: str = traceback_msg

    def is_error(self) -> bool:
        return True

    def get_err(self) -> Exception:
        if self.traceback_msg is not None:
            return type(self.err)(self.traceback_msg)
        return self.err


def worker_function(
    worker: 'ParallelWorkersWrapper',
    context: CX,
    use_savers: bool,
    logging_queue: Optional[Queue],
):
    """
    Worker function that transforms the items inside the queue of items to process.

    :param queue_worker: step to transform
    :param context: execution context
    :param use_savers: use savers
    :return:
    """
    try:
        if use_savers:
            worker.reload_post_saving(context)
        step = worker.get_step()

        register_log_producer_for_main_logger_thread_to_consume(logging_queue)

        while True:
            task: QueuedMinibatchTask = None
            try:
                task = worker._get_minibatch_to_consume()
                if task.is_error():
                    break
                task.minibatch_dact = step.handle_transform(
                    task.minibatch_dact, context)

                if not isinstance(task.minibatch_dact, DACT):
                    raise ValueError(
                        f"Minibatch DACT is not of good type. Received {type( task.minibatch_dact)}: {str( task.minibatch_dact)}.")
                    pickle.dumps(task)

            except queue.Empty:
                # Queue did a timeout, continuing.
                task = None
                continue  # breakpoint here.
            except _QueueDestroyedError:
                # This error can happen at the destruction of workers or processes.
                task = None
                return  # breakpoint here.
            except Exception as err:
                context.flow.log_error(err)
                task.minibatch_dact = None
                task = pickle_exception_into_task(task, err)
                pass
            finally:
                if task is not None:
                    worker.put_minibatch_produced_to_next_consumers(task)
    except Exception as err:
        context.flow.log_error(err)
        task = pickle_exception_into_task(task, err)
        worker.put_minibatch_produced_to_next_consumers(task)
    else:
        pass  # This pass is for breakpoints to be set if needed.
    finally:
        pass  # This pass is for breakpoints to be set if needed.


def pickle_exception_into_task(task, err) -> MinibatchError:
    if task is None:
        task = QueuedMinibatchTask(None, None)
    return task.to_error(err)


class ParallelWorkersWrapper(_ProducerConsumerMixin, MetaStep):
    """
    Start multiple Process or Thread that consumes items from the minibatch DACT Queue, and produces them on the next registered consumers' queue.
    """

    def __init__(
        self,
        wrapped: BaseTransformer,
        max_queued_minibatches: int = None,
        n_workers: int = 1,
        use_processes: bool = True,
        use_savers: bool = False
    ):
        MetaStep.__init__(self, wrapped)
        _ProducerConsumerMixin.__init__(self, max_queued_minibatches)
        self.n_workers: int = n_workers
        self.use_processes: bool = use_processes
        self.use_savers = use_savers

        self.running_workers: List[Process] = []
        # TODO: maybe start this thread in the main instead than in every worker's main.
        self.logging_thread: Optional[ParallelLoggingConsumerThread] = None

    def _setup(self, context: 'CX' = None) -> Optional[RecursiveDict]:
        MetaStep._setup(self)
        _ProducerConsumerMixin._setup(self)

    def __getstate__(self):
        """
        This class, upon being forked() to a new process with pickles,
        should not copy references to other threads or processes.
        """
        state = self.__dict__.copy()
        state['running_workers'] = []  # delete self.workers in the process fork's pickling.
        state['logging_thread'] = None
        return state

    def start(self, context: CX, logging_queue: Optional[Queue] = None):
        """
        Start multiple processes or threads with the worker function as a target.
        These workers will consume minibatches from the queue and produce them on the next queue(s).
        They are started as multiprocessing daemons, so that they will not block the main
        process if there is an error requiring to exit.

        :param context: An execution context that will be checked to be thread_safe.
        :param logging_queue: An optional logging_queue from the object :class:`ParallelLoggingConsumerThread` to pass and recover parallelized log records to. Not required for thread-only parallelism, only process-based parallelism.
        """
        if self.input_queue is None:
            raise ValueError("Please call self._setup before and connect queues.")
        if self.use_savers:
            _ = self.save(context, full_dump=True)  # Cannot delete queue worker self.
            del self.wrapped  # that will be reloaded in the new thread or process.

        process_safe_context: CX = context
        process_safe_context = process_safe_context.thread_safe()
        ParallelObj = Thread
        if self.use_processes:  # TODO: that was to debug pickles.
            # New process requires trimming the references to other processes
            # when we create many processes: https://stackoverflow.com/a/65749012
            # Important check to avoid this cPython issue to cause a deadlock: https://github.com/python/cpython/issues/79423
            process_safe_context = process_safe_context.process_safe()
            ParallelObj = Process

        if self.use_processes and logging_queue is None:
            self.logging_thread = ParallelLoggingConsumerThread()
            logging_queue = self.logging_thread.logging_queue
            self.logging_thread.start()

        self.running_workers = []
        for i in range(self.n_workers):
            p = ParallelObj(
                target=worker_function,
                name=self.name + f"__worker_function{i}",
                args=(self, process_safe_context, self.use_savers, logging_queue)
            )
            p.daemon = True
            self.running_workers.append(p)
            p.start()

        if self.use_savers:
            self.reload_post_saving(context)

    def reload_post_saving(self, context: CX) -> 'ParallelWorkersWrapper':
        self._assert(
            self.use_savers,
            "Attempted to reload post saving whilst not using savers."
            "Don't reload when not using savers.",
            context)
        saved_worker: ParallelWorkersWrapper = context.load(self.get_name())
        self.set_step(saved_worker.get_step())

    def join(self):
        """
        Wait for workers to finish at least their capture of the logging calls.
        """
        _ProducerConsumerMixin.join(self)
        if self.logging_thread is not None:
            self.logging_thread.join(timeout=5.0)
            self.logging_thread = None

    def _teardown(self) -> Optional[RecursiveDict]:
        """
        Stop all processes on teardown.
        """
        self.stop()
        _ProducerConsumerMixin._teardown(self)
        return MetaStep._teardown(self)

    def stop(self):
        """
        Stop all of the workers.

        :return:
        """
        if self.use_processes:
            for w in self.running_workers:
                w.terminate()
        self.join()
        self.running_workers = []
        self.consumers = []


QueuedPipelineStepsTuple = Union[
    BaseTransformer,  # step
    Tuple[int, BaseTransformer],  # (n_workers, step)
    Tuple[str, BaseTransformer],  # (step_name, step)
    Tuple[str, int, BaseTransformer],  # (step_name, n_workers, step)
    Tuple[str, int, int, BaseTransformer],  # (step_name, n_workers, max_queued_minibatches, step)
]
FullQueuedPipelineStepsTuple = Tuple[str, int, int, List[Tuple], BaseTransformer]


class BaseQueuedPipeline(MiniBatchSequentialPipeline):
    """
    Sub class of :class:`Pipeline`.
    Transform data in many pipeline steps at once in parallel in the pipeline using multiprocessing Queues.

    Example usage :

    .. code-block:: python

        # Multiple ways of specifying the steps tuples exists to do various things:
        # step name, step
        p = SequentialQueuedPipeline([
            ('step_a', Identity()),
            ('step_b', Identity()),
        ], n_workers=1, batch_size=10, max_queued_minibatches=10)

        # step name, number of workers, step
        p = SequentialQueuedPipeline([
            ('step_a', 1, Identity()),
            ('step_b', 1, Identity()),
        ], batch_size=10, max_queued_minibatches=10)

        # step name, number of workers, and max size
        p = SequentialQueuedPipeline([
            ('step_a', 1, 10, Identity()),
            ('step_b', 1, 10, Identity()),
        ], batch_size=10)

        # step name, number of workers for each step, and additional argument for each worker
        p = SequentialQueuedPipeline([
            ('step_a', 1, [('host', 'host1'), ('host', 'host2')], 10, Identity())
        ], batch_size=10)

        # step name, number of workers for each step, additional argument for each worker, and max size
        p = SequentialQueuedPipeline([
            ('step_a', 1, [('host', 'host1'), ('host', 'host2')], 10, Identity())
        ], batch_size=10)

        # It's also possible to do parallel feature unions:
        n_workers = 4
        worker_arguments = [('hyperparams', HyperparameterSamples({'multiply_by': 2})) for _ in range(n_workers)]
        p = ParallelQueuedFeatureUnion([
            ('1', n_workers, worker_arguments, MultiplyByN()),
        ], batch_size=10, max_queued_minibatches=5)
        outputs = p.transform(list(range(100)))

    :param steps: pipeline steps.
    :param batch_size: number of elements to combine into a single batch.
    :param n_workers_per_step: number of workers to spawn per step.
    :param max_queued_minibatches: max number of batches inside the processing queue between the workers.
    :param data_joiner: transformer step to join streamed batches together at the end of the pipeline.
    :param use_processes: use processes instead of threads for parallel processing. multiprocessing.Process is used by default.
    :param use_savers: use savers to serialize steps for parallel processing. Recommended if using processes instead of threads.
    :param keep_incomplete_batch: (Optional.) A bool that indicates whether
    or not the last batch should be dropped in the case it has fewer than
    `batch_size` elements; the default behavior is to keep the smaller batch.
    :param default_value_data_inputs: expected_outputs default fill value
    for padding and values outside iteration range, or :class:`~neuraxle.data_container.StripAbsentValues`
    to trim absent values from the batch
    :param default_value_expected_outputs: expected_outputs default fill value
    for padding and values outside iteration range, or :class:`~neuraxle.data_container.StripAbsentValues`
    to trim absent values from the batch
    """

    def __init__(
            self,
            steps: List[QueuedPipelineStepsTuple],
            batch_size: int,
            n_workers_per_step: int = None,
            max_queued_minibatches: int = None,
            data_joiner: BaseTransformer = None,
            use_processes: bool = False,
            use_savers: bool = False,
            keep_incomplete_batch: bool = True,
            default_value_data_inputs: Union[Any, StripAbsentValues] = None,
            default_value_expected_outputs: Union[Any, StripAbsentValues] = None,
    ):
        self.batch_size: int = batch_size
        self.n_workers_per_step: int = n_workers_per_step
        self.max_queued_minibatches: int = max_queued_minibatches or 0
        if data_joiner is None:
            # Note that the data joiner differs from the workers joiner, in the sense that the
            # data joiner is applied after the data is joined by the workers joiner.
            data_joiner = NumpyConcatenateOuterBatch()
        self.data_joiner: BaseTransformer = data_joiner
        self.use_processes: bool = use_processes
        self.use_savers: bool = use_savers
        self.keep_incomplete_batch: bool = keep_incomplete_batch
        self.default_value_data_inputs: Union[Any, StripAbsentValues] = default_value_data_inputs
        self.default_value_expected_outputs: Union[Any, StripAbsentValues] = default_value_expected_outputs

        self.is_pipeline_connected = False

        MiniBatchSequentialPipeline.__init__(
            self,
            steps=self._parallel_wrap_all_steps_tuples(steps),
            batch_size=batch_size,
            keep_incomplete_batch=keep_incomplete_batch,
            default_value_data_inputs=default_value_data_inputs,
            default_value_expected_outputs=default_value_expected_outputs
        )
        self._refresh_steps()

        self.logging_thread: Optional[ParallelLoggingConsumerThread] = None

    def _parallel_wrap_all_steps_tuples(self, all_step_tuples: NamedStepsList) -> NamedStepsList:
        """
        Wrap each step by a :class:`QueueWorker` to  allow data to flow in many pipeline steps at once in parallel.

        :param steps: (name, n_workers, step)
        :type steps: NameNWorkerStepTupleList
        :return: steps as tuple
        """
        steps_as_tuple: NamedStepsList = []
        for step_tuple in all_step_tuples:
            parallel_worker_wrapper = self._parallel_wrap_step_tuple(step_tuple)
            steps_as_tuple.append((parallel_worker_wrapper.name, parallel_worker_wrapper))

        steps_as_tuple.append(("WorkersJoiner", WorkersJoiner(batch_size=self.batch_size)))
        return steps_as_tuple

    def _parallel_wrap_step_tuple(self, step_tuple: QueuedPipelineStepsTuple):
        name, n_workers, max_queued_minibatches, actual_step = self._parse_step_tuple(
            step_tuple)

        return ParallelWorkersWrapper(
            actual_step,
            n_workers=n_workers,
            use_processes=self.use_processes,
            max_queued_minibatches=max_queued_minibatches,
            use_savers=self.use_savers
        ).set_name(name)

    def _parse_step_tuple(self, step_tuple: QueuedPipelineStepsTuple) -> FullQueuedPipelineStepsTuple:
        """
        Return all params necessary to create the QueuedPipeline for the given step.

        :param step_tuple: the un-parsed tuple of steps
        :type step: QueuedPipelineStepsTupleList
        :return: a tuple of (name, n_workers, max_queued_minibatches, actual_step)
        :rtype: Tuple[str, int, int, BaseStep]
        """
        if isinstance(step_tuple, BaseTransformer):
            step_tuple = (step_tuple,)
        elif len(step_tuple) > 5:
            raise Exception(f'Invalid Queued Pipeline Steps Shape: {step_tuple}.')
        actual_step: BaseTransformer = step_tuple[-1]

        # Default values before parse, if missing:
        name: str = actual_step.name
        n_workers: int = max(1, self.n_workers_per_step or 1)
        max_queued_minibatches: int = self.max_queued_minibatches

        if len(step_tuple) == 2:
            if isinstance(step_tuple[0], str):
                name = step_tuple[0]
            else:
                n_workers = step_tuple[0]
        elif len(step_tuple) >= 3:
            name = step_tuple[0]
            n_workers = step_tuple[1]
            if len(step_tuple) == 4:
                max_queued_minibatches = step_tuple[2]

        return (name, n_workers, max_queued_minibatches, actual_step)

    def _will_process(
        self, data_container: DACT, context: CX
    ) -> Tuple[DACT, CX]:
        """
        Setup streaming pipeline before any handler methods.

        :param data_container: data container
        :param context: execution context
        :return:
        """
        self._setup(context=context)
        context = context.synchroneous()
        return data_container.copy(), context

    def _setup(self, context: CX = None) -> 'BaseTransformer':
        """
        Connect the queued workers together so that the data can correctly flow through the pipeline.

        :param context: execution context
        :return: step
        :rtype: BaseStep
        """
        workers_joiner: WorkersJoiner = self.joiner
        workers_joiner._setup(context=context)
        for step in self.body:
            step: ParallelWorkersWrapper = step
            step._setup(context=context)

        self._connect_queued_pipeline()

        return MiniBatchSequentialPipeline._setup(self, context=context)

    def fit_transform_data_container(
        self, data_container: DACT, context: CX
    ) -> Tuple[Pipeline, DACT]:
        """
        Fit transform sequentially if any step is fittable, such as with :class:`MiniBatchSequentialPipeline`. Otherwise transform in parallel as it should.

        :param data_container: data container
        :type data_container: DataContainer
        :param context: execution context
        :type context: ExecutionContext
        :return:
        """
        has_fittable_step: bool = False
        for step in self.body:
            if isinstance(step.get_step(), _FittableStep) and not isinstance(step.get_step(), NonFittableMixin):
                has_fittable_step = True

        if has_fittable_step:
            # will use MiniBatchSequentialPipeline parent:
            self.is_invalidated = True
            return super().fit_transform_data_container(data_container, context)
        else:
            # Use the parallel transform without fitting:
            data_container = self.transform_data_container(data_container, context)
            data_container = self._did_transform(data_container, context)
            return self, data_container

    def transform_data_container(self, data_container: DACT, context: CX) -> DACT:
        """
        Transform data container

        :param data_container: data container to transform.
        :type data_container: DataContainer
        :param context: execution context
        :type context: ExecutionContext
        :return: data container
        """
        logging_queue: Optional[Queue] = None
        if self.use_processes:
            self.logging_thread = ParallelLoggingConsumerThread()
            logging_queue = self.logging_thread.logging_queue
            self.logging_thread.start()

        for step in self.values():
            if step.input_queue is None:
                raise ValueError("Please connect queues and do the self._setup before attempting to join.")

        # start steps with parallelized context.
        for step in self.body:
            step: ParallelWorkersWrapper = step
            step.start(context, logging_queue)

        # prepare minibatch iterator:
        data_container.set_ids(data_container.ids)
        minibatch_iterator: Iterable[DACT[IDT, DIT, EOT]] = data_container.minibatches(
            batch_size=self.batch_size,
            keep_incomplete_batch=self.keep_incomplete_batch,
            default_value_data_inputs=self.default_value_data_inputs,
            default_value_expected_outputs=self.default_value_expected_outputs
        )

        # send batches to input queues and label queue output expected summaries.
        n_minibatches_per_worker = 0
        for minibatch_index, minibatch_dact in enumerate(minibatch_iterator):
            task = QueuedMinibatchTask(minibatch_dact, self.name)

            self._dispatch_minibatch_to_consumer_workers(
                minibatch_index=minibatch_index, task=task)

            n_minibatches_per_worker += 1

        # join output queues.
        n_workers = self.get_n_workers_to_join()
        workers_joiner: WorkersJoiner = self.joiner
        workers_joiner.set_join_quantities(n_workers, n_minibatches_per_worker)
        data_container = workers_joiner.join_workers(data_container, context)

        # for step in self.body:
        #     step: ParallelWorkersWrapper = step
        #     step.join()

        return data_container

    def _did_transform(self, data_container: DACT, context: CX) -> DACT:
        """
        Stop all of the workers after transform. Also, join the data using self.data_joiner.

        :param data_container: data container
        :type data_container: DataContainer
        :param context: execution context
        :type context: ExecutionContext
        :return: data container
        :rtype: DataContainer
        """
        for step in self.body:
            step: ParallelWorkersWrapper = step
            step.stop()
        if self.logging_thread is not None:
            self.logging_thread.join(timeout=5.0)
            self.logging_thread = None

        return self.data_joiner.handle_transform(data_container, context)

    def _did_process(self, data_container: PredsDACT, context: CX) -> PredsDACT:
        self._disconnect_queued_pipeline()
        return super()._did_process(data_container, context)

    @abstractmethod
    def get_n_workers_to_join(self) -> int:
        """
        Get the total number of terminal steps at the end of each row of queued workers.
        """
        raise NotImplementedError()

    @abstractmethod
    def _connect_queued_pipeline(self):
        """
        Connect all the queued workers together so that the data can flow through each step.

        :return:
        """
        if not self.is_pipeline_connected:
            # register queues and summaries in workers_joiner...
            self.is_pipeline_connected = True
        raise NotImplementedError()

    def _disconnect_queued_pipeline(self):
        if self.is_pipeline_connected:
            workers_joiner: WorkersJoiner = self.joiner
            workers_joiner.join()
            for step in self.body:
                step: ParallelWorkersWrapper = step
                step.join()
        self.is_pipeline_connected = False

    @abstractmethod
    def _dispatch_minibatch_to_consumer_workers(self, minibatch_index: int, task: QueuedMinibatchTask):
        """
        Send batches to queued pipeline. It is blocking if there is no more space available in the multiprocessing queues.
        Workers might return batches in a different order, but the queue joiner will reorder them at the end.
        The queue joiner will use the summary ids to reorder all of the received batches.

        :param batch_index: batch index
        :param data_container: data container batch
        :return:
        """
        raise NotImplementedError()


class SequentialQueuedPipeline(BaseQueuedPipeline):
    """
    Using :class:`QueueWorker`, run all steps sequentially even if they are in separate processes or threads.
    This is a parallel pipeline that uses a queue to communicate between the steps, and which parallelizes the steps
    using many workers in different processes or threads.

    This pipeline is useful when the steps are independent of each other and can be run in parallel. This is especially
    the case when the steps are not fittable, such as inheriting from the :class:`~neuraxle.base.NonFittableMixin`.
    Otherwise, fitting may not be parallelized, although the steps can be run in parallel for the transformation.

    .. seealso::
        :class:`~neuraxle.pipeline.BasePipeline`,
        :func:`~neuraxle.data_container.DataContainer.minibatches`,
        :class:`~neuraxle.data_container.StripAbsentValues`,
        :class:`QueueWorker`,
        :class:`BaseQueuedPipeline`,
        :class:`ParallelQueuedPipeline`,
    """

    def get_n_workers_to_join(self) -> int:
        return 1

    def _connect_queued_pipeline(self):
        """
        Sequentially connect of the queued workers as producers and consumers.

        :return:
        """
        if not self.is_pipeline_connected:
            for i, (name, consumer) in enumerate(self[1:]):
                producer: _ProducerConsumerMixin = self[i]
                producer.register_consumer(consumer)
            self.is_pipeline_connected = True

    def _dispatch_minibatch_to_consumer_workers(self, minibatch_index: int, task: QueuedMinibatchTask):
        """
        Send batches to process to the first queued worker.

        :param batch_index: batch index
        :param data_container: data container batch
        :return:
        """
        workers_joiner: WorkersJoiner = self.joiner

        # TODO: extract method.
        last_consumer_name: str = self.steps_as_tuple[-2][0]
        workers_joiner.append_terminal_summary(last_consumer_name, task)

        first_consumer: _ProducerConsumerMixin = self[0]
        first_consumer.put_minibatch_produced(task)


class ParallelQueuedFeatureUnion(BaseQueuedPipeline):
    """
    Using :class:`QueueWorker`, run all steps in parallel using QueueWorkers.

    .. seealso::
        :class:`QueueWorker`,
        :class:`BaseQueuedPipeline`,
        :class:`SequentialQueuedPipeline`,
    """

    def get_n_workers_to_join(self):
        n_worker_wrappers = len(self) - 1
        return n_worker_wrappers

    def _connect_queued_pipeline(self):
        """
        Connect the queue joiner to all of the queued workers to process data in parallel.

        :return:
        """
        if not self.is_pipeline_connected:
            joiner_consumer: _ProducerConsumerMixin = self.joiner
            for producer in self.body:
                producer: _ProducerConsumerMixin = producer
                producer.register_consumer(joiner_consumer)
            self.is_pipeline_connected = True

    def _dispatch_minibatch_to_consumer_workers(self, minibatch_index: int, task: QueuedMinibatchTask):
        """
        In the case of the feature union, sending a batch to the workers is done by sending the batch
        to each of the workers that will work in parallel to consume the same copy sent to all.
        """
        # TODO: task index in task.
        workers_joiner: WorkersJoiner = self.joiner
        for consumer in self.body:
            consumer: ParallelWorkersWrapper = consumer
            workers_joiner.append_terminal_summary(consumer.get_name(), task)
            consumer.put_minibatch_produced(task)


MiniBatchDataContainer = ListDataContainer[IDT, DIT, EOT]
SummaryIDStr = str  # Picture this as an IDT where each ID is a summary ID
MetaListDataContainer = ListDataContainer[
    List[SummaryIDStr],
    MiniBatchDataContainer,
    List[None]
]


class WorkersJoiner(_ProducerConsumerMixin, Joiner):
    """
    Consume the results of the other :class:`_ProducerConsumerMixin` workers to join their data.
    Also do error handling.
    """

    def __init__(self, batch_size: int, n_worker_wrappers_to_join: int = None):
        # TODO: the present class could be renamed to WorkersTerminalConsumer or so. The [-1] item of the pipeline should be the data joiner instead.
        Joiner.__init__(self, batch_size=batch_size)
        _ProducerConsumerMixin.__init__(self)
        self.n_workers = n_worker_wrappers_to_join
        self.names: List[str] = []
        self.summaries: List[str] = []

    def _setup(self, context: 'CX' = None) -> Optional[RecursiveDict]:
        _ProducerConsumerMixin._setup(self, context)
        return Joiner._setup(self, context)

    def _teardown(self) -> Optional[RecursiveDict]:
        """
        Properly clean queue, summary ids, and results during teardown.

        :return: teardowned self
        """
        self.names = []
        self.summaries = []
        _ProducerConsumerMixin._teardown(self)
        return Joiner._teardown(self)

    def set_join_quantities(self, n_workers: int, n_minibatches_per_worker: int):
        self.n_workers = n_workers
        self.n_workers_remaining = n_workers
        self.n_minibatches_per_worker = n_minibatches_per_worker
        self.remaining_batches_per_workers: Dict[str, int] = defaultdict(lambda: self.n_minibatches_per_worker)

    def append_terminal_summary(self, name: str, task: QueuedMinibatchTask):
        """
        Append the summary id of the worker to the list of summaries.

        :param name: name of the worker
        :param task: task
        :return:
        """
        self.names.append(name)
        self.summaries.append(task.minibatch_dact.get_ids_summary())

    def join_workers(self, original_dact: DACT, sync_context: CX) -> DACT:
        """
        Return the accumulated results of the workers.

        :return: transformed data container
        :rtype: DataContainer
        """
        if self.input_queue is None:
            raise ValueError("Please connect queues and do the self._setup before attempting to join.")

        if self.n_minibatches_per_worker == 0:
            return original_dact

        # Gather minibatch tasks to consume and organize them per step name:
        step_to_minibatches_dacts: Dict[str, MetaListDataContainer] = \
            self._consume_enqueued_minibatches(sync_context)

        # Sort step_to_minibatches_dacts by step_name index in the step names list as like in the loop below but based on the order of steps. In the case of a pipeline, there will be only the last step anyway so the sort will return without even sorting. For feature union, there are many.
        list_dact: List[ListDataContainer] = \
            self._merge_minibatches(step_to_minibatches_dacts)

        _ProducerConsumerMixin.join(self)

        return original_dact.set_data_inputs(list_dact)

    def _consume_enqueued_minibatches(self, sync_context) -> Dict[str, MetaListDataContainer]:

        step_to_minibatches_dacts: Dict[str, MetaListDataContainer] = \
            defaultdict(MiniBatchDataContainer.empty)
        # shape: [worker, minibatches, dact_of_a_minibatch]
        # The shape is confusing because the ListDataContainer contains itself some other ListDataContainer.

        while self.n_workers_remaining > 0:
            # Note: the call here is most likely the reason for
            #       deadlocks if something weird happened in the worker_function.

            try:
                task: QueuedMinibatchTask = self._get_minibatch_to_consume()  # breakpoint here.

                if task.is_error():
                    task: MinibatchError = task
                    err = task.get_err()
                    raise err from err
                else:
                    step_to_minibatches_dacts[task.worker_name].append_data_container_in_data_inputs(
                        task.minibatch_dact)

                    self.remaining_batches_per_workers[task.worker_name] -= 1
                    if self.remaining_batches_per_workers[task.worker_name] == 0:
                        self.n_workers_remaining -= 1
            except queue.Empty:
                continue

        return step_to_minibatches_dacts

    def _merge_minibatches(self, step_to_minibatches_dacts: Dict[str, MetaListDataContainer]) -> List[ListDataContainer]:
        step_to_minibatches_dacts = OrderedDict(sorted(
            step_to_minibatches_dacts.items(),
            key=lambda str_dact_tup: self.names.index(str_dact_tup[0])
        ))

        # Join all the results based on step name:
        list_dact: List[ListDataContainer] = []  # shape: [worker, regular_dact]

        for minibatches_list_dacts in step_to_minibatches_dacts.values():
            minibatches_list_dacts: MetaListDataContainer = minibatches_list_dacts

            # reorder results by ids of summary
            # TODO: check if could use original dact instead. or other batch count int number, that would be more efficient.
            minibatches_list_dacts.data_inputs.sort(key=lambda dc: self.summaries.index(dc.get_ids_summary()))

            sorted_step_dact = ListDataContainer.empty()
            for minibatch_dact in minibatches_list_dacts.data_inputs:
                minibatch_dact: MiniBatchDataContainer = minibatch_dact
                sorted_step_dact.extend(minibatch_dact)

            list_dact.append(sorted_step_dact)

        return list_dact



"""
Neuraxle's Flask Wrapper classes
====================================
The flask wrapper classes are used to easily serve pipeline predictions using a flask rest api.

..
    Copyright 2019, Neuraxio Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

..
    Thanks to Umaneo Technologies Inc. for their contributions to this Machine Learning
    project, visit https://www.umaneo.com/ for more information on Umaneo Technologies Inc.

"""
from abc import ABC, abstractmethod

import numpy as np
from flask import Response

from neuraxle.base import BaseStep
from neuraxle.pipeline import Pipeline


class JSONDataBodyDecoder(BaseStep, ABC):
    """
    Class to be used within a FlaskRESTApiWrapper to convert input json to actual data (e.g.: arrays)
    """

    def transform(self, data_inputs):
        return self.decode(data_inputs)

    @abstractmethod
    def decode(self, data_inputs: dict):
        """
        Will convert data_inputs to a dict or a compatible data structure for jsonification

        :param encoded data_inputs (dict parsed from json):
        :return: data_inputs (as a data structure compatible with pipeline's data inputs)
        """
        raise NotImplementedError("TODO: inherit from the `JSONDataBodyDecoder` class and implement this method.")


class JSONDataResponseEncoder(BaseStep, ABC):
    """
    Base class to be used within a FlaskRESTApiWrapper to convert prediction output to json response.
    """

    def transform(self, data_inputs) -> Response:
        """
        Transform processed data inputs into a flask response object.

        :param data_inputs:
        :return: flask response object
        """
        from flask import jsonify
        return jsonify(self.encode(data_inputs))

    @abstractmethod
    def encode(self, data_inputs) -> dict:
        """
        Convert data_inputs to a dict or a compatible data structure for jsonification.

        :param data_inputs (a data structure outputted by the pipeline after a transform):
        :return: encoded data_inputs (jsonifiable dict)
        """
        raise NotImplementedError("TODO: inherit from the `JSONDataResponseEncoder` class and implement this method.")


class FlaskRestApiWrapper(Pipeline):
    """
    Wrap a pipeline to easily deploy it to a REST API. Just provide a json encoder and a json decoder.

    Usage example:

    ```
    class CustomJSONDecoderFor2DArray(JSONDataBodyDecoder):
        '''This is a custom JSON decoder class that precedes the pipeline's transformation.'''

        def decode(self, data_inputs: dict):
            values_in_json_2d_arr: List[List[int]] = data_inputs["values"]
            return np.array(values_in_json_2d_arr)

    class CustomJSONEncoderOfOutputs(JSONDataResponseEncoder):
        '''This is a custom JSON response encoder class for converting the pipeline's transformation outputs.'''

        def encode(self, data_inputs) -> dict:
            return {
                'predictions': list(data_inputs)
            }

    app = FlaskRestApiWrapper(
        json_decoder=CustomJSONDecoderFor2DArray(),
        wrapped=Pipeline(...),
        json_encoder=CustomJSONEncoderOfOutputs(),
    ).get_app()

    app.run(debug=False, port=5000)
    ```
    """

    def __init__(
            self,
            json_decoder: JSONDataBodyDecoder,
            wrapped: BaseStep,
            json_encoder: JSONDataResponseEncoder,
            route='/'):
        Pipeline.__init__(self, [
            json_decoder,
            wrapped,
            json_encoder
        ])
        self.route: str = route

    def get_app(self):
        """
        This methods returns a REST API wrapping the pipeline.

        :return: a Flask app (as given by `app = Flask(__name__)` and then configured).
        """
        from flask import Flask, request
        from flask_restful import Api, Resource

        app = Flask(__name__)
        api = Api(app)
        wrapped = self

        class RESTfulRes(Resource):
            def get(self):
                return wrapped.transform(request.get_json())

        api.add_resource(
            RESTfulRes,
            self.route
        )

        return app



"""
Output Handlers Steps
====================================
You can find here output handlers steps that changes especially the data outputs.

..
    Copyright 2019, Neuraxio Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

..
    Thanks to Umaneo Technologies Inc. for their contributions to this Machine Learning
    project, visit https://www.umaneo.com/ for more information on Umaneo Technologies Inc.

"""
import copy
from typing import Tuple

from neuraxle.base import BaseStep
from neuraxle.base import ExecutionContext as CX
from neuraxle.base import (ForceHandleOnlyMixin, MetaStep,
                           MixinForBaseTransformer)
from neuraxle.data_container import DataContainer as DACT


class OutputTransformerWrapper(ForceHandleOnlyMixin, MetaStep):
    """
    A step that can sends the expected_outputs to the wrapped step
    so that it can transform the expected outputs.
    """

    def __init__(self, wrapped):
        MetaStep.__init__(self, wrapped)
        ForceHandleOnlyMixin.__init__(self)

    def _transform_data_container(self, data_container: DACT, context: CX) -> DACT:
        """
        Handle transform by passing expected outputs to the wrapped step transform method.
        Update the expected outputs with the outputs.

        :param context: execution context
        :param data_container:
        :return: data container
        :rtype: DataContainer
        """
        new_expected_outputs_data_container = self.wrapped.handle_transform(
            DACT(
                ids=data_container.ids,
                data_inputs=data_container.eo,
                expected_outputs=None
            ),
            context
        )
        self._set_expected_outputs(data_container, new_expected_outputs_data_container, context)

        return data_container

    def _fit_data_container(
        self, data_container: DACT, context: CX
    ) -> Tuple[BaseStep, DACT]:
        """
        Handle fit by passing expected outputs to the wrapped step fit method.

        :param context: execution context
        :type context: ExecutionContext
        :param data_container: data container to fit on
        :return: self, data container
        """
        self.wrapped = self.wrapped.handle_fit(
            DACT(
                data_inputs=data_container.expected_outputs,
                ids=data_container.ids,
                expected_outputs=None
            ),
            context
        )

        return self, data_container

    def _fit_transform_data_container(
        self, data_container: DACT, context: CX
    ) -> Tuple[BaseStep, DACT]:
        """
        Handle fit transform by passing expected outputs to the wrapped step fit method.
        Update the expected outputs with the outputs.

        :param context: execution context
        :type context: ExecutionContext
        :param data_container: data container to fit on
        :return: self, data container
        """
        self.wrapped, new_expected_outputs_data_container = self.wrapped.handle_fit_transform(
            DACT(
                data_inputs=data_container.expected_outputs,
                ids=data_container.ids,
                expected_outputs=None
            ),
            context
        )
        self._set_expected_outputs(data_container, new_expected_outputs_data_container, context)

        return self, data_container

    def handle_inverse_transform(self, data_container: DACT, context: CX) -> DACT:
        """
        Handle inverse transform by passing expected outputs to the wrapped step inverse transform method.
        Update the expected outputs with the outputs.

        :param context: execution context
        :param data_container:
        :return: data container
        :rtype: DataContainer
        """
        new_expected_outputs_data_container = self.wrapped.handle_inverse_transform(
            DACT(
                data_inputs=data_container.expected_outputs,
                ids=data_container.ids,
                expected_outputs=None
            ),
            context.push(self.wrapped)
        )

        self._set_expected_outputs(data_container, new_expected_outputs_data_container, context)

        return data_container

    def _set_expected_outputs(
            self, data_container: DACT, new_expected_outputs_data_container: DACT, context: CX
    ) -> DACT:

        self._assert(
            len(data_container) == len(new_expected_outputs_data_container),
            'OutputTransformerWrapper: Found different len for old data inputs, and expected outputs '
            'to reinsert. Please return the same the same amount of data inputs, and expected outputs, '
            'or otherwise create your own handler methods to do more funky things.',
            context
        )

        data_container.set_expected_outputs(new_expected_outputs_data_container.data_inputs)
        data_container.set_ids(new_expected_outputs_data_container._ids)

        return data_container


class _DidProcessIdsInputOutputHandlerMixin(MixinForBaseTransformer):
    def _did_process(self, data_container: DACT, context: CX) -> DACT:
        ids, di, eo = data_container.data_inputs
        self._assert(
            di is None or eo is None or ids is None or (len(di) == len(eo) and len(ids) == len(di)),
            f'{self.name}: Found different len for non-null data inputs, and expected outputs. '
            f'Please return the same the same amount of data inputs, and expected outputs, or '
            f'otherwise create your own handler methods to do more funky things.',
            context
        )

        data_container.set_ids(ids=ids)
        data_container.set_data_inputs(data_inputs=di)
        data_container.set_expected_outputs(expected_outputs=eo)

        data_container = super()._did_process(data_container, context)

        self._assert(
            len(data_container.ids) == len(data_container.data_inputs),
            f'{self.name}: Caching broken because there is a different len of current ids, and data inputs. Please use InputAndOutputTransformerWrapper if you plan to change the len of the data inputs.',
            context
        )

        return data_container


class IdsInputAndOutputTransformerWrapper(_DidProcessIdsInputOutputHandlerMixin, ForceHandleOnlyMixin, MetaStep):
    """
    Wrapper step to transform both ids, data inputs, and expected output at the same time using classical fit and transform methods.
    It sends the data_inputs, and the expected_outputs to the wrapped step so that it can transform them.

    .. seealso::
        :class:`~neuraxle.base.BaseStep`,
        :class:`~neuraxle.base.ForceHandleOnlyMixin`
    """

    def __init__(self, wrapped):
        MetaStep.__init__(self, wrapped)
        ForceHandleOnlyMixin.__init__(self)
        _DidProcessIdsInputOutputHandlerMixin.__init__(self)

    def _transform_data_container(self, data_container: DACT, context: CX) -> DACT:
        """
        Handle transform by passing data_inputs, and expected outputs to the wrapped step transform method.
        Update the expected outputs with the outputs.

        :param context: execution context
        :param data_container:
        :return: data container
        :rtype: DataContainer
        """
        output_data_container = self.wrapped.handle_transform(
            DACT(
                ids=data_container._ids,
                data_inputs=(data_container._ids, data_container.data_inputs, data_container.expected_outputs),
                expected_outputs=None
            ),
            context
        )

        return output_data_container

    def _fit_data_container(
        self, data_container: DACT, context: CX
    ) -> Tuple[BaseStep, DACT]:
        """
        Handle fit by passing the data inputs, and the expected outputs to the wrapped step fit method.

        :param context: execution context
        :type context: ExecutionContext
        :param data_container: data container to fit on
        :return: self, data container
        """
        self.wrapped = self.wrapped.handle_fit(
            DACT(
                ids=data_container._ids,
                data_inputs=(copy.copy(data_container._ids), copy.copy(
                    data_container.data_inputs), copy.copy(data_container.expected_outputs)),
                expected_outputs=None
            ),
            context
        )

        data_container.set_data_inputs(
            (data_container._ids, data_container.data_inputs, data_container.expected_outputs))
        data_container.set_expected_outputs(expected_outputs=None)  # TODO: thy this?

        return self

    def _fit_transform_data_container(
        self, data_container: DACT, context: CX
    ) -> Tuple[BaseStep, DACT]:
        """
        Handle fit transform by passing the data inputs, and the expected outputs to the wrapped step fit method.
        Update the expected outputs with the outputs.

        :param context: execution context
        :type context: ExecutionContext
        :param data_container: data container to fit on
        :return: self, data container
        """
        self.wrapped, output_data_container = self.wrapped.handle_fit_transform(
            DACT(
                ids=data_container._ids,
                data_inputs=(data_container._ids, data_container.data_inputs, data_container.expected_outputs),
                expected_outputs=None
            ),
            context
        )
        return self, output_data_container

    def handle_inverse_transform(self, data_container: DACT, context: CX) -> DACT:
        """
        Handle inverse transform by passing the data inputs, and the expected outputs to the wrapped step inverse transform method.
        Update the expected outputs with the outputs.

        :param context: execution context
        :param data_container:
        :return: data container
        :rtype: DataContainer
        """
        output_data_container = self.wrapped.handle_inverse_transform(
            DACT(
                ids=data_container._ids,
                data_inputs=(data_container._ids, data_container.data_inputs, data_container.expected_outputs),
                expected_outputs=None
            ),
            context.push(self.wrapped)
        )

        return output_data_container


class IdsAndInputAndOutputTransformerMixin(_DidProcessIdsInputOutputHandlerMixin):
    """
    Base output transformer step that can modify ids, data inputs, and expected_outputs at the same time.
    """

    def _transform_data_container(self, data_container: DACT, context: CX) -> DACT:
        """
        Handle inverse transform by updating the data inputs, and expected outputs inside the data container.

        :param context: execution context
        :param data_container:
        :return:
        """
        di_eo = (data_container._ids, data_container.data_inputs, data_container.expected_outputs)
        new_ids, new_data_inputs, new_expected_outputs = self.transform(di_eo)
        data_container.set_data_inputs((new_ids, new_data_inputs, new_expected_outputs))
        return data_container

    def _fit_data_container(self, data_container: DACT, context: CX) -> 'BaseStep':
        """
        Handle transform by fitting the step,
        and updating the data inputs, and expected outputs inside the data container.

        :param context: execution context
        :param data_container:
        :return:
        """
        new_self = self.fit((data_container._ids, data_container.data_inputs, data_container.expected_outputs), None)
        data_container.set_data_inputs(
            (data_container._ids, data_container.data_inputs, data_container.expected_outputs))
        data_container.set_expected_outputs(expected_outputs=None)  # TODO: check if this None eo is correct
        return new_self

    def _fit_transform_data_container(
        self, data_container: DACT, context: CX
    ) -> Tuple['BaseStep', DACT]:
        """
        Handle transform by fitting the step,
        and updating the data inputs, and expected outputs inside the data container.

        :param context: execution context
        :param data_container:
        :return:
        """
        new_self, (new_ids, new_data_inputs, new_expected_outputs) = self.fit_transform(
            (data_container._ids, data_container.data_inputs, data_container.expected_outputs), None)
        data_container.set_data_inputs((new_ids, new_data_inputs, new_expected_outputs))
        return new_self, data_container
